################################################################################
# Module: archetypal.template
# Description:
# License: MIT, see full license in LICENSE.txt
# Web: https://github.com/samuelduchesne/archetypal
################################################################################

import collections
import logging as lg
import math
import sqlite3
from enum import Enum

import pandas as pd
from deprecation import deprecated
from sigfig import round

import archetypal
from archetypal import log, settings, timeit
from archetypal.template import UmiBase, UmiSchedule, UniqueName
from archetypal.utils import reduce
import numpy as np


class DimmingTypes(Enum):
    Continuous = 0
    Off = 1
    Stepped = 2

    def __lt__(self, other):
        return self._value_ < other._value_

    def __gt__(self, other):
        return self._value_ > other._value_


class ZoneLoad(UmiBase):
    """Zone Loads

    Important:
        Please note that the calculation of the equipment power density will sum
        up the electric equipment objects as well as the gas equipment objects.

    .. image:: ../images/template/zoneinfo-loads.png
    """

    def __init__(
        self,
        DimmingType=DimmingTypes.Continuous,
        EquipmentAvailabilitySchedule=None,
        EquipmentPowerDensity=0,
        IlluminanceTarget=500,
        LightingPowerDensity=0,
        LightsAvailabilitySchedule=None,
        OccupancySchedule=None,
        IsEquipmentOn=True,
        IsLightingOn=True,
        IsPeopleOn=True,
        PeopleDensity=0,
        **kwargs,
    ):
        """Initialize a new ZoneLoad object

        Args:
            DimmingType (int): Different types to dim the lighting to respect the
                IlluminanceTarget and taking into account the daylight illuminance:
                    - Continuous = 0, the overhead lights dim continuously and
                      linearly from (maximum electric power, maximum light output) to (
                      minimum electric power, minimum light output) as the daylight
                      illuminance increases. The lights stay on at the minimum point
                      with further increase in the daylight illuminance.
                    - Off = 1, Lights switch off completely when the minimum
                      dimming point is reached.
                    - Stepped = 2, the electric power input and light output vary
                      in discrete, equally spaced steps.
            EquipmentAvailabilitySchedule (UmiSchedule): The name of
                the schedule (Day | Week | Year) that modifies the design level
                parameter for electric equipment.
            EquipmentPowerDensity (float): Equipment Power Density in the zone
                (W/m²).
            IlluminanceTarget (float): Number of lux to be respected in the zone
            LightingPowerDensity (float): Lighting Power Density in the zone
                (W/m²).
            LightsAvailabilitySchedule (UmiSchedule): The name of the
                schedule (Day | Week | Year) that modifies the design level
                parameter for lighting.
            OccupancySchedule (UmiSchedule): The name of the schedule
                (Day | Week | Year) that modifies the number of people parameter
                for electric equipment.
            IsEquipmentOn (bool): If True, heat gains from Equipment are taken
                into account for the zone's load calculation.
            IsLightingOn (bool): If True, heat gains from Lights are taken into
                account for the zone's load calculation.
            IsPeopleOn (bool): If True, heat gains from People are taken into
                account for the zone's load calculation.
            PeopleDensity (float): Density of people in the zone (people/m²).
            **kwargs: Other keywords passed to the parent constructor :class:`UmiBase`.
        """
        super(ZoneLoad, self).__init__(**kwargs)
        self.DimmingType = DimmingTypes(DimmingType)
        self.EquipmentAvailabilitySchedule = EquipmentAvailabilitySchedule
        self.EquipmentPowerDensity = EquipmentPowerDensity
        self.IlluminanceTarget = IlluminanceTarget
        self.LightingPowerDensity = LightingPowerDensity
        self.LightsAvailabilitySchedule = LightsAvailabilitySchedule
        self.OccupancySchedule = OccupancySchedule
        self.IsEquipmentOn = IsEquipmentOn
        self.IsLightingOn = IsLightingOn
        self.IsPeopleOn = IsPeopleOn
        self.PeopleDensity = PeopleDensity

        self._belongs_to_zone = kwargs.get("zone", None)

    @property
    def EquipmentPowerDensity(self):
        if math.isnan(self._EquipmentPowerDensity):
            return self._EquipmentPowerDensity
        return round(float(self._EquipmentPowerDensity), decimals=3)

    @EquipmentPowerDensity.setter
    def EquipmentPowerDensity(self, value):
        self._EquipmentPowerDensity = value

    @property
    def IlluminanceTarget(self):
        return float(self._IlluminanceTarget)

    @IlluminanceTarget.setter
    def IlluminanceTarget(self, value):
        self._IlluminanceTarget = value

    @property
    def LightingPowerDensity(self):
        if math.isnan(self._LightingPowerDensity):
            return self._LightingPowerDensity
        return round(float(self._LightingPowerDensity), decimals=3)

    @LightingPowerDensity.setter
    def LightingPowerDensity(self, value):
        self._LightingPowerDensity = value

    @property
    def PeopleDensity(self):
        if math.isnan(self._PeopleDensity):
            return self._PeopleDensity
        return round(float(self._PeopleDensity), decimals=3)

    @PeopleDensity.setter
    def PeopleDensity(self, value):
        self._PeopleDensity = value

    def __add__(self, other):
        """
        Args:
            other (Zone):
        """
        return self.combine(other)

    def __hash__(self):
        return hash(
            (self.__class__.__name__, getattr(self, "Name", None), self.DataSource)
        )

    def __eq__(self, other):
        if not isinstance(other, ZoneLoad):
            return NotImplemented
        else:
            return all(
                [
                    self.DimmingType == other.DimmingType,
                    self.EquipmentAvailabilitySchedule
                    == other.EquipmentAvailabilitySchedule,
                    self.EquipmentPowerDensity == other.EquipmentPowerDensity,
                    self.IlluminanceTarget == other.IlluminanceTarget,
                    self.LightingPowerDensity == other.LightingPowerDensity,
                    self.LightsAvailabilitySchedule == other.LightsAvailabilitySchedule,
                    self.OccupancySchedule == other.OccupancySchedule,
                    self.IsEquipmentOn == other.IsEquipmentOn,
                    self.IsLightingOn == other.IsLightingOn,
                    self.IsPeopleOn == other.IsPeopleOn,
                    self.PeopleDensity == other.PeopleDensity,
                ]
            )

    @classmethod
    @deprecated(
        deprecated_in="1.3.1",
        removed_in="1.5",
        current_version=archetypal.__version__,
        details="Use from_dict function instead",
    )
    def from_json(cls, *args, **kwargs):

        return cls.from_dict(*args, **kwargs)

    @classmethod
    def from_dict(cls, *args, **kwargs):
        """
        Args:
            *args:
            **kwargs:
        """
        zl = cls(*args, **kwargs)

        cool_schd = kwargs.get("EquipmentAvailabilitySchedule", None)
        zl.EquipmentAvailabilitySchedule = zl.get_ref(cool_schd)
        heat_schd = kwargs.get("LightsAvailabilitySchedule", None)
        zl.LightsAvailabilitySchedule = zl.get_ref(heat_schd)
        mech_schd = kwargs.get("OccupancySchedule", None)
        zl.OccupancySchedule = zl.get_ref(mech_schd)

        return zl

    def to_json(self):
        """Convert class properties to dict"""
        self.validate()  # Validate object before trying to get json format

        data_dict = collections.OrderedDict()

        data_dict["$id"] = str(self.id)
        data_dict["DimmingType"] = self.DimmingType.value
        data_dict[
            "EquipmentAvailabilitySchedule"
        ] = self.EquipmentAvailabilitySchedule.to_dict()
        data_dict["EquipmentPowerDensity"] = round(self.EquipmentPowerDensity, 3)
        data_dict["IlluminanceTarget"] = round(self.IlluminanceTarget, 3)
        data_dict["LightingPowerDensity"] = round(self.LightingPowerDensity, 3)
        data_dict[
            "LightsAvailabilitySchedule"
        ] = self.LightsAvailabilitySchedule.to_dict()
        data_dict["OccupancySchedule"] = self.OccupancySchedule.to_dict()
        data_dict["IsEquipmentOn"] = self.IsEquipmentOn
        data_dict["IsLightingOn"] = self.IsLightingOn
        data_dict["IsPeopleOn"] = self.IsPeopleOn
        data_dict["PeopleDensity"] = round(self.PeopleDensity, 3)
        data_dict["Category"] = self.Category
        data_dict["Comments"] = self.Comments
        data_dict["DataSource"] = self.DataSource
        data_dict["Name"] = UniqueName(self.Name)

        return data_dict

    @classmethod
    @timeit
    def from_zone(cls, zone, **kwargs):
        """
        Args:
            zone (ZoneDefinition): zone to gets information from
        """
        # If Zone is not part of total area, it should not have a ZoneLoad object.
        if not zone._is_part_of_total_floor_area:
            return None

        # Get schedule index for different loads and create ZoneLoad arguments
        # Verify if Equipment in zone

        # create database connection with sqlite3
        with sqlite3.connect(str(zone.idf.sql_file)) as conn:
            sql_query = "select ifnull(ZoneIndex, null) from Zones where ZoneName=?"
            t = (zone.Name.upper(),)
            c = conn.cursor()
            c.execute(sql_query, t)
            (zone_index,) = c.fetchone()

            sql_query = "select t.* from NominalElectricEquipment t where ZoneIndex=?"
            nominal_elec = pd.read_sql(sql_query, conn, params=(zone_index,))

            sql_query = "select t.* from NominalGasEquipment t where ZoneIndex=?"
            nominal_gas = pd.read_sql(sql_query, conn, params=(zone_index,))

            def get_schedule(series):
                """Computes schedule with quantity for nominal equipment
                series"""
                sched = series["ScheduleIndex"]
                sql_query = (
                    "select t.ScheduleName, t.ScheduleType as M from "
                    "Schedules t where ScheduleIndex=?"
                )
                sched_name, sched_type = c.execute(sql_query, (int(sched),)).fetchone()
                level_ = float(series["DesignLevel"])
                if level_ > 0:
                    return UmiSchedule(
                        Name=sched_name,
                        idf=zone.idf,
                        Type=sched_type,
                        quantity=level_,
                    )

            schedules = []
            if not nominal_elec.empty:
                # compute schedules series
                elec_scds = nominal_elec.apply(get_schedule, axis=1).to_list()
                elec_scds = list(filter(None, elec_scds))
                schedules.extend(elec_scds)

            if not nominal_gas.empty:
                # compute schedules series
                gas_scds = nominal_gas.apply(get_schedule, axis=1).to_list()
                gas_scds = list(filter(None, gas_scds))
                schedules.extend(gas_scds)

            if schedules:
                EquipmentAvailabilitySchedule = reduce(
                    UmiSchedule.combine,
                    schedules,
                    quantity=True,
                )
                EquipmentPowerDensity = (
                    EquipmentAvailabilitySchedule.quantity / zone.area
                )
            else:
                EquipmentAvailabilitySchedule = None
                EquipmentPowerDensity = np.NaN

            # Verifies if Lights in zone
            sql_query = "select t.* from NominalLighting t where ZoneIndex=?"
            nominal_lighting = pd.read_sql(sql_query, conn, params=(zone_index,))

            lighting_schedules = []
            if not nominal_lighting.empty:
                # compute schedules series
                light_scds = nominal_lighting.apply(get_schedule, axis=1)
                lighting_schedules.extend(light_scds)

            if lighting_schedules:
                LightsAvailabilitySchedule = reduce(
                    UmiSchedule.combine,
                    lighting_schedules,
                    quantity=True,
                )
                LightingPowerDensity = LightsAvailabilitySchedule.quantity / zone.area
            else:
                LightsAvailabilitySchedule = None
                LightingPowerDensity = np.NaN

            # Verifies if People in zone

            def get_schedule(series):
                """Computes schedule with quantity for nominal equipment
                series"""
                sched = series["NumberOfPeopleScheduleIndex"]
                sql_query = (
                    "select t.ScheduleName, t.ScheduleType as M from "
                    "Schedules t where ScheduleIndex=?"
                )
                sched_name, sched_type = c.execute(sql_query, (int(sched),)).fetchone()
                return UmiSchedule(
                    Name=sched_name,
                    idf=zone.idf,
                    Type=sched_type,
                    quantity=series["NumberOfPeople"],
                )

            sql_query = "select t.* from NominalPeople t where ZoneIndex=?"
            nominal_people = pd.read_sql(sql_query, conn, params=(zone_index,))

            occupancy_schedules = []
            if not nominal_people.empty:
                # compute schedules series
                occ_scds = nominal_people.apply(get_schedule, axis=1)
                occupancy_schedules.extend(occ_scds)

            if occupancy_schedules:
                OccupancySchedule = reduce(
                    UmiSchedule.combine,
                    occupancy_schedules,
                    quantity=lambda x: sum(obj.quantity for obj in x),
                )
                PeopleDensity = OccupancySchedule.quantity / zone.area
            else:
                OccupancySchedule = None
                PeopleDensity = np.NaN

        name = zone.Name + "_ZoneLoad"
        z_load = cls(
            Name=name,
            zone=zone,
            DimmingType=_resolve_dimming_type(zone),
            EquipmentAvailabilitySchedule=EquipmentAvailabilitySchedule,
            EquipmentPowerDensity=EquipmentPowerDensity,
            IlluminanceTarget=_resolve_illuminance_target(zone),
            LightingPowerDensity=LightingPowerDensity,
            LightsAvailabilitySchedule=LightsAvailabilitySchedule,
            OccupancySchedule=OccupancySchedule,
            IsEquipmentOn=EquipmentPowerDensity > 0,
            IsLightingOn=LightingPowerDensity > 0,
            IsPeopleOn=PeopleDensity > 0,
            PeopleDensity=PeopleDensity,
            idf=zone.idf,
            Category=zone.idf.name,
            **kwargs,
        )
        return z_load

    def combine(self, other, weights=None):
        """Combine two ZoneLoad objects together.

        Args:
            other (ZoneLoad):
            weights (list-like, optional): A list-like object of len 2. If None,
                the volume of the zones for which self and other belongs is
                used.

        Returns:
            (ZoneLoad): the combined ZoneLoad object.
        """
        # Check if other is None. Simply return self
        if not other:
            return self

        if not self:
            return other

        # Check if other is the same type as self
        if not isinstance(other, self.__class__):
            msg = "Cannot combine %s with %s" % (
                self.__class__.__name__,
                other.__class__.__name__,
            )
            raise NotImplementedError(msg)

        # Check if other is not the same as self
        if self == other:
            return self

        incoming_load_data = self.__dict__.copy()
        incoming_load_data.pop("Name")

        meta = self._get_predecessors_meta(other)

        if not weights:
            zone_weight = settings.zone_weight
            weights = [
                getattr(self._belongs_to_zone, str(zone_weight)),
                getattr(other._belongs_to_zone, str(zone_weight)),
            ]
            log(
                'using zone {} "{}" as weighting factor in "{}" '
                "combine.".format(
                    zone_weight,
                    " & ".join(list(map(str, map(int, weights)))),
                    self.__class__.__name__,
                )
            )

        new_attr = dict(
            DimmingType=max(self.DimmingType, other.DimmingType),
            EquipmentAvailabilitySchedule=UmiSchedule.combine(
                self.EquipmentAvailabilitySchedule,
                other.EquipmentAvailabilitySchedule,
                quantity=True,
            ),
            EquipmentPowerDensity=self._float_mean(
                other, "EquipmentPowerDensity", weights
            ),
            IlluminanceTarget=self._float_mean(other, "IlluminanceTarget", weights),
            LightingPowerDensity=self._float_mean(
                other, "LightingPowerDensity", weights
            ),
            LightsAvailabilitySchedule=UmiSchedule.combine(
                self.LightsAvailabilitySchedule,
                other.LightsAvailabilitySchedule,
                quantity=True,
            ),
            OccupancySchedule=UmiSchedule.combine(
                self.OccupancySchedule,
                other.OccupancySchedule,
                quantity=True,
            ),
            IsEquipmentOn=any([self.IsEquipmentOn, other.IsEquipmentOn]),
            IsLightingOn=any([self.IsLightingOn, other.IsLightingOn]),
            IsPeopleOn=any([self.IsPeopleOn, other.IsPeopleOn]),
            PeopleDensity=self._float_mean(other, "PeopleDensity", weights),
        )

        new_obj = self.__class__(
            **meta, **new_attr, idf=self.idf, allow_duplicates=self._allow_duplicates
        )
        new_obj._belongs_to_zone = self._belongs_to_zone
        new_obj.predecessors.update(self.predecessors + other.predecessors)
        return new_obj

    def validate(self):
        """Validate object and fill in missing values."""
        if not self.DimmingType:
            self.DimmingType = DimmingTypes.Continuous
        if not self.EquipmentAvailabilitySchedule:
            self.EquipmentAvailabilitySchedule = UmiSchedule.constant_schedule()
        if not self.EquipmentPowerDensity:
            self.EquipmentPowerDensity = 0
        if not self.IlluminanceTarget:
            self.IlluminanceTarget = 500
        if not self.LightingPowerDensity:
            self.LightingPowerDensity = 0
        if not self.LightsAvailabilitySchedule:
            self.LightsAvailabilitySchedule = UmiSchedule.constant_schedule()
        if not self.OccupancySchedule:
            self.OccupancySchedule = UmiSchedule.constant_schedule()
        if not self.IsEquipmentOn:
            self.IsEquipmentOn = False
        if not self.IsLightingOn:
            self.IsLightingOn = False
        if not self.IsPeopleOn:
            self.IsPeopleOn = False
        if not self.PeopleDensity:
            self.PeopleDensity = 0
        return self

    def mapping(self):
        self.validate()

        return dict(
            DimmingType=self.DimmingType,
            EquipmentAvailabilitySchedule=self.EquipmentAvailabilitySchedule,
            EquipmentPowerDensity=self.EquipmentPowerDensity,
            IlluminanceTarget=self.IlluminanceTarget,
            LightingPowerDensity=self.LightingPowerDensity,
            LightsAvailabilitySchedule=self.LightsAvailabilitySchedule,
            OccupancySchedule=self.OccupancySchedule,
            IsEquipmentOn=self.IsEquipmentOn,
            IsLightingOn=self.IsLightingOn,
            IsPeopleOn=self.IsPeopleOn,
            PeopleDensity=self.PeopleDensity,
            Category=self.Category,
            Comments=self.Comments,
            DataSource=self.DataSource,
            Name=self.Name,
        )

    def get_ref(self, ref):
        """Get item matching reference id.

        Args:
            ref:
        """
        return next(
            iter(
                [value for value in ZoneLoad.CREATED_OBJECTS if value.id == ref["$ref"]]
            ),
            None,
        )


def _resolve_dimming_type(zone):
    """Resolves the dimming type for the Zone object"""
    # First, retrieve the list of Daylighting objects for this zone. Uses the eppy
    # `getreferingobjs` method.
    ep_obj = zone._epbunch
    possible_ctrls = ep_obj.getreferingobjs(
        iddgroups=["Daylighting"], fields=["Zone_Name"]
    )
    # Then, if there are controls
    if possible_ctrls:
        # Filter only the "Daylighting:Controls"
        ctrls = [
            ctrl
            for ctrl in possible_ctrls
            if ctrl.key.upper() == "Daylighting:Controls".upper()
        ]
        ctrl_types = [ctrl["Lighting_Control_Type"] for ctrl in ctrls]

        # There should only be one control per zone. A set of controls should return 1.
        if len(set(ctrl_types)) == 1:
            dimming_type, *_ = set(ctrl_types)
            if dimming_type.lower() not in ["continuous", "stepped"]:
                raise ValueError(
                    f"A dimming type of type '{dimming_type}' for zone '{zone.Name}' is not yet supported in UMI"
                )
            else:
                log(f"Dimming type for zone '{zone.Name}' set to '{dimming_type}'")
                return DimmingTypes[dimming_type]  # Return first element
        else:
            raise ValueError(
                "Could not resolve more than one dimming types for Zone {}. "
                "Make sure there is only one".format(zone.Name)
            )
    else:
        # Else, there are no dimming controls => set to "Off".
        log(
            "No dimming type found for zone {}. Setting as Off".format(zone.Name),
            lg.DEBUG,
        )
        return DimmingTypes.Off


def _resolve_illuminance_target(zone):
    """Resolves the illuminance target for the Zone object"""
    # First, retrieve the list of Daylighting objects for this zone. Uses the eppy
    # `getreferingobjs` method.
    ep_obj = zone._epbunch
    possible_ctrls = ep_obj.getreferingobjs(
        iddgroups=["Daylighting"], fields=["Zone_Name"]
    )
    # Then, if there are controls
    if possible_ctrls:
        # Filter only the "Daylighting:Controls"
        ctrls = [
            ctrl
            for ctrl in possible_ctrls
            if ctrl.key.upper() == "Daylighting:Controls".upper()
        ]
        ctrl_types = [
            ctrl["Illuminance_Setpoint_at_Reference_Point_1"] for ctrl in ctrls
        ]

        # There should only be one control per zone. A set of controls should return 1.
        if len(set(ctrl_types)) == 1:
            dimming_type = next(iter(set(ctrl_types)))
            log(f"Illuminance target for zone '{zone.Name}' set to '{dimming_type}'")
            return float(dimming_type)  # Return first element
        else:
            raise ValueError(
                "Could not resolve more than one illuminance targets for Zone {}. "
                "Make sure there is only one".format(zone.Name)
            )
    else:
        # Else, there are no dimming controls => set to "Off".
        log(
            "No illuminance target found for zone {}. Setting to default 500 "
            "lux".format(zone.Name),
            lg.DEBUG,
        )
        return 500



"""ExpandObjects module"""

import logging as lg
import platform
import shutil
import subprocess
import time
from io import StringIO
from subprocess import CalledProcessError
from threading import Thread

from eppy.runner.run_functions import paths_from_version
from path import Path
from tqdm import tqdm

from archetypal.eplus_interface.energy_plus import EnergyPlusProgram
from archetypal.eplus_interface.exceptions import EnergyPlusVersionError
from archetypal.eplus_interface.version import EnergyPlusVersion
from archetypal.utils import log


class ExpandObjectsExe(EnergyPlusProgram):
    """ExpandObject Wrapper"""

    def __init__(self, idf):
        """Constructor."""
        super().__init__(idf)

    @property
    def cmd(self):
        """Get the command."""
        return ["ExpandObjects"]


class ExpandObjectsThread(Thread):
    """ExpandObjects program manager."""

    def __init__(self, idf, tmp):
        """Constructor."""
        super(ExpandObjectsThread, self).__init__()
        self.p = None
        self.std_out = None
        self.std_err = None
        self.idf = idf
        self.cancelled = False
        self.run_dir = Path("")
        self.exception = None
        self.name = "ExpandObjects_" + self.idf.name
        self.tmp = tmp
        self.cmd = None

    def run(self):
        """Wrapper around the EnergyPlus command line interface."""
        try:
            self.cancelled = False
            # get version from IDF object or by parsing the IDF file for it

            # Move files into place
            tmp = self.tmp
            self.epw = self.idf.epw.copy(tmp / "in.epw").expand()
            self.idfname = Path(self.idf.savecopy(tmp / "in.idf")).expand()
            self.idd = self.idf.iddname.copy(tmp / "Energy+.idd").expand()
            self.expandobjectsexe = Path(
                shutil.which("ExpandObjects", path=self.eplus_home.expand())
            ).copy2(tmp)
            self.run_dir = Path(tmp).expand()

            # Run ExpandObjects Program
            self.cmd = ExpandObjectsExe(self.idf).cmd
            with tqdm(
                unit_scale=True,
                miniters=1,
                desc=f"ExpandObjects #{self.idf.position}-{self.idf.name}",
                position=self.idf.position,
            ) as progress:

                self.p = subprocess.Popen(
                    self.cmd,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    shell=True,  # can use shell
                    cwd=self.run_dir.abspath(),
                )
                start_time = time.time()
                # self.msg_callback("ExpandObjects started")
                for line in self.p.stdout:
                    self.msg_callback(line.decode("utf-8"))
                    progress.update()

                # We explicitly close stdout
                self.p.stdout.close()

                # Wait for process to complete
                self.p.wait()

                # Communicate callbacks
                if self.cancelled:
                    self.msg_callback("ExpandObjects cancelled")
                    # self.cancelled_callback(self.std_out, self.std_err)
                else:
                    if self.p.returncode == 0:
                        self.msg_callback(
                            f"ExpandObjects completed in "
                            f"{time.time() - start_time:,.2f} seconds"
                        )
                        self.success_callback()
                    else:
                        self.failure_callback()
        except Exception as e:
            self.exception = e
            self.p.kill()  # kill process to be sure
            return

    def msg_callback(self, *args, **kwargs):
        """Pass message to logger."""
        log(*args, name=self.idf.name, **kwargs)

    def success_callback(self):
        """Replace idf with expanded.idf."""
        if (self.run_dir / "expanded.idf").exists():
            if isinstance(self.idf.idfname, StringIO):
                file = StringIO((self.run_dir / "expanded.idf").read_text())
            else:
                file = (self.run_dir / "expanded.idf").copy(self.idf.idfname)
            self.idf.idfname = file
        if (Path(self.run_dir) / "GHTIn.idf").exists():
            self.idf.include.append(
                (Path(self.run_dir) / "GHTIn.idf").copy(
                    self.idf.output_directory / "GHTIn.idf"
                )
            )

        if (Path(self.run_dir) / "BasementGHTIn.idf").exists():
            self.idf.include.append(
                (Path(self.run_dir) / "BasementGHTIn.idf").copy(
                    self.idf.output_directory / "BasementGHTIn.idf"
                )
            )

    def failure_callback(self):
        """Read stderr and pass to logger."""
        for line in self.p.stderr:
            self.msg_callback(line.decode("utf-8"), level=lg.ERROR)
        self.exception = CalledProcessError(
            self.p.returncode, cmd=self.cmd, stderr=self.p.stderr
        )

    def cancelled_callback(self, stdin, stdout):
        """Call on cancelled."""
        pass

    @property
    def eplus_home(self):
        eplus_exe, eplus_home = paths_from_version(self.idf.as_version.dash)
        if not Path(eplus_home).exists():
            raise EnergyPlusVersionError(
                msg=f"No EnergyPlus Executable found for version "
                f"{EnergyPlusVersion(self.idf.as_version)}"
            )
        else:
            return Path(eplus_home)



import collections
import logging as lg

from deprecation import deprecated

from archetypal import __version__, log, reduce, timeit
from archetypal.template import OpaqueConstruction, UmiBase, UniqueName


class ZoneConstructionSet(UmiBase):
    """Zone-specific :class:`Construction` ids"""

    def __init__(
        self,
        Name,
        Facade=None,
        Ground=None,
        Partition=None,
        Roof=None,
        Slab=None,
        IsFacadeAdiabatic=False,
        IsGroundAdiabatic=False,
        IsPartitionAdiabatic=False,
        IsRoofAdiabatic=False,
        IsSlabAdiabatic=False,
        **kwargs,
    ):
        """
        Args:
            Name (str): Name of the object. Must be Unique.
            Facade (OpaqueConstruction): The OpaqueConstruction object representing
                a facade.
            Ground (OpaqueConstruction): The OpaqueConstruction object representing
                a ground floor.
            Partition (OpaqueConstruction): The OpaqueConstruction object representing
                a partition wall.
            Roof (OpaqueConstruction): The OpaqueConstruction object representing
                a roof.
            Slab (OpaqueConstruction): The OpaqueConstruction object representing
                a slab.
            IsFacadeAdiabatic (bool): If True, surface is adiabatic.
            IsGroundAdiabatic (bool): If True, surface is adiabatic.
            IsPartitionAdiabatic (bool): If True, surface is adiabatic.
            IsRoofAdiabatic (bool): If True, surface is adiabatic.
            IsSlabAdiabatic (bool): If True, surface is adiabatic.
            **kwargs:
        """
        super(ZoneConstructionSet, self).__init__(Name, **kwargs)
        self.Slab = Slab
        self.IsSlabAdiabatic = IsSlabAdiabatic
        self.Roof = Roof
        self.IsRoofAdiabatic = IsRoofAdiabatic
        self.Partition = Partition
        self.IsPartitionAdiabatic = IsPartitionAdiabatic
        self.Ground = Ground
        self.IsGroundAdiabatic = IsGroundAdiabatic
        self.Facade = Facade
        self.IsFacadeAdiabatic = IsFacadeAdiabatic
        self._belongs_to_zone = kwargs.get("zone", None)

    def __add__(self, other):
        """Overload + to implement self.combine.

        Args:
            other:
        """
        return self.combine(other)

    def __hash__(self):
        return hash(
            (self.__class__.__name__, getattr(self, "Name", None), self.DataSource)
        )

    def __eq__(self, other):
        if not isinstance(other, ZoneConstructionSet):
            return NotImplemented
        else:
            return all(
                [
                    self.Slab == other.Slab,
                    self.IsSlabAdiabatic == other.IsSlabAdiabatic,
                    self.Roof == other.Roof,
                    self.IsRoofAdiabatic == other.IsRoofAdiabatic,
                    self.Partition == other.Partition,
                    self.IsPartitionAdiabatic == other.IsPartitionAdiabatic,
                    self.Ground == other.Ground,
                    self.IsGroundAdiabatic == other.IsGroundAdiabatic,
                    self.Facade == other.Facade,
                    self.IsFacadeAdiabatic == other.IsFacadeAdiabatic,
                ]
            )

    @classmethod
    @timeit
    def from_zone(cls, zone, **kwargs):
        """
        Args:
            zone (ZoneDefinition):
        """
        name = zone.Name + "_ZoneConstructionSet"
        # dispatch surfaces
        facade, ground, partition, roof, slab = [], [], [], [], []
        zonesurfaces = zone._zonesurfaces
        for surf in zonesurfaces:
            for disp_surf in surface_dispatcher(surf, zone):
                if disp_surf:
                    if disp_surf.Surface_Type == "Facade":
                        if zone.is_part_of_conditioned_floor_area:
                            facade.append(disp_surf)
                    elif disp_surf.Surface_Type == "Ground":
                        ground.append(disp_surf)
                    elif disp_surf.Surface_Type == "Partition":
                        partition.append(disp_surf)
                    elif disp_surf.Surface_Type == "Roof":
                        roof.append(disp_surf)
                    elif disp_surf.Surface_Type == "Slab":
                        slab.append(disp_surf)
                    else:
                        msg = (
                            'Surface Type "{}" is not known, this method is not'
                            " implemented".format(disp_surf.Surface_Type)
                        )
                        raise NotImplementedError(msg)

        # Returning a set() for each groups of Constructions.

        facades = set(facade)
        if facades:
            facade = reduce(OpaqueConstruction.combine, facades)
        else:
            facade = None
        grounds = set(ground)
        if grounds:
            ground = reduce(OpaqueConstruction.combine, grounds)
        else:
            ground = None
        partitions = set(partition)
        if partitions:
            partition = reduce(OpaqueConstruction.combine, partitions)
        else:
            partition = None
        roofs = set(roof)
        if roofs:
            roof = reduce(OpaqueConstruction.combine, roofs)
        else:
            roof = None
        slabs = set(slab)
        if slabs:
            slab = reduce(OpaqueConstruction.combine, slabs)
        else:
            slab = None

        z_set = cls(
            Facade=facade,
            Ground=ground,
            Partition=partition,
            Roof=roof,
            Slab=slab,
            Name=name,
            zone=zone,
            idf=zone.idf,
            Category=zone.idf.name,
            **kwargs,
        )
        return z_set

    @classmethod
    @deprecated(
        deprecated_in="1.3.1",
        removed_in="1.5",
        current_version=__version__,
        details="Use from_dict function instead",
    )
    def from_json(cls, *args, **kwargs):
        return cls.from_dict(*args, **kwargs)

    @classmethod
    def from_dict(cls, *args, **kwargs):
        """
        Args:
            *args:
            **kwargs:
        """
        zc = cls(*args, **kwargs)

        ref = kwargs.get("Facade", None)
        zc.Facade = zc.get_ref(ref)

        ref = kwargs.get("Ground", None)
        zc.Ground = zc.get_ref(ref)

        ref = kwargs.get("Partition", None)
        zc.Partition = zc.get_ref(ref)

        ref = kwargs.get("Roof", None)
        zc.Roof = zc.get_ref(ref)

        ref = kwargs.get("Slab", None)
        zc.Slab = zc.get_ref(ref)

        return zc

    def combine(self, other, weights=None, **kwargs):
        """Append other to self. Return self + other as a new object.

        Args:
            other (ZoneConstructionSet):
            weights:

        Returns:
            (ZoneConstructionSet): the combined ZoneConstructionSet object.
        """
        # Check if other is None. Simply return self
        if not other:
            return self

        if not self:
            return other
        # Check if other is the same type as self
        if not isinstance(other, self.__class__):
            msg = "Cannot combine %s with %s" % (
                self.__class__.__name__,
                other.__class__.__name__,
            )
            raise NotImplementedError(msg)

        # Check if other is not the same as self
        if self == other:
            return self

        if not weights:
            weights = [self._belongs_to_zone.volume, other._belongs_to_zone.volume]
            log(
                'using zone volume "{}" as weighting factor in "{}" '
                "combine.".format(
                    " & ".join(list(map(str, map(int, weights)))),
                    self.__class__.__name__,
                )
            )

        meta = self._get_predecessors_meta(other)
        new_attr = dict(
            Slab=OpaqueConstruction.combine(self.Slab, other.Slab),
            IsSlabAdiabatic=any([self.IsSlabAdiabatic, other.IsSlabAdiabatic]),
            Roof=OpaqueConstruction.combine(self.Roof, other.Roof),
            IsRoofAdiabatic=any([self.IsRoofAdiabatic, other.IsRoofAdiabatic]),
            Partition=OpaqueConstruction.combine(self.Partition, other.Partition),
            IsPartitionAdiabatic=any(
                [self.IsPartitionAdiabatic, other.IsPartitionAdiabatic]
            ),
            Ground=OpaqueConstruction.combine(self.Ground, other.Ground),
            IsGroundAdiabatic=any([self.IsGroundAdiabatic, other.IsGroundAdiabatic]),
            Facade=OpaqueConstruction.combine(self.Facade, other.Facade),
            IsFacadeAdiabatic=any([self.IsFacadeAdiabatic, other.IsFacadeAdiabatic]),
        )
        new_obj = self.__class__(**meta, **new_attr, idf=self.idf, **kwargs)
        new_obj.predecessors.update(self.predecessors + other.predecessors)
        return new_obj

    def to_json(self):
        """Convert class properties to dict"""
        self.validate()
        data_dict = collections.OrderedDict()

        data_dict["$id"] = str(self.id)
        data_dict["Facade"] = {"$ref": str(self.Facade.id)}
        data_dict["Ground"] = {"$ref": str(self.Ground.id)}
        data_dict["Partition"] = {"$ref": str(self.Partition.id)}
        data_dict["Roof"] = {"$ref": str(self.Roof.id)}
        data_dict["Slab"] = {"$ref": str(self.Slab.id)}
        data_dict["IsFacadeAdiabatic"] = self.IsFacadeAdiabatic
        data_dict["IsGroundAdiabatic"] = self.IsGroundAdiabatic
        data_dict["IsPartitionAdiabatic"] = self.IsPartitionAdiabatic
        data_dict["IsRoofAdiabatic"] = self.IsRoofAdiabatic
        data_dict["IsSlabAdiabatic"] = self.IsSlabAdiabatic
        data_dict["Category"] = self.Category
        data_dict["Comments"] = self.Comments
        data_dict["DataSource"] = self.DataSource
        data_dict["Name"] = UniqueName(self.Name)

        return data_dict

    def validate(self):
        for attr in ["Slab", "Roof", "Partition", "Ground", "Facade"]:
            if getattr(self, attr) is None:
                # First try to get one from another zone that has the attr
                zone = next(
                    iter(
                        filter(
                            lambda x: getattr(x, attr, None) is not None,
                            UmiBase.CREATED_OBJECTS,
                        )
                    ),
                    None,
                )
                if zone:
                    setattr(self, attr, getattr(zone, attr))
                else:
                    # If not, default to a generic construction for last resort.
                    setattr(self, attr, OpaqueConstruction.generic(idf=self.idf))
                log(
                    f"While validating {self}, the required attribute "
                    f"'{attr}' was filled "
                    f"with {getattr(self, attr)}",
                    lg.DEBUG,
                )
        return self

    @staticmethod
    def _do_facade(surf):
        """
        Args:
            surf (EpBunch):
        """
        log(
            'surface "%s" assigned as a Facade' % surf.Name,
            lg.DEBUG,
            name=surf.theidf.name,
        )
        oc = OpaqueConstruction.from_epbunch(
            surf.theidf.getobject("Construction".upper(), surf.Construction_Name)
        )
        oc.area = surf.area
        oc.Surface_Type = "Facade"
        oc.Category = oc.Surface_Type
        return oc

    @staticmethod
    def _do_ground(surf):
        """
        Args:
            surf (EpBunch):
        """
        log(
            'surface "%s" assigned as a Ground' % surf.Name,
            lg.DEBUG,
            name=surf.theidf.name,
        )
        oc = OpaqueConstruction.from_epbunch(
            surf.theidf.getobject("Construction".upper(), surf.Construction_Name)
        )
        oc.area = surf.area
        oc.Surface_Type = "Ground"
        oc.Category = oc.Surface_Type
        return oc

    @staticmethod
    def _do_partition(surf):
        """
        Args:
            surf (EpBunch):
        """
        the_construction = surf.theidf.getobject(
            "Construction".upper(), surf.Construction_Name
        )
        if the_construction:
            oc = OpaqueConstruction.from_epbunch(the_construction)
            oc.area = surf.area
            oc.Surface_Type = "Partition"
            oc.Category = oc.Surface_Type
            log(
                'surface "%s" assigned as a Partition' % surf.Name,
                lg.DEBUG,
                name=surf.theidf.name,
            )
            return oc
        else:
            # we might be in a situation where the construction does not exist in the
            # file. For example, this can happen when the construction is defined as
            # "Air Wall", which is a construction type internal to EnergyPlus.
            return None

    @staticmethod
    def _do_roof(surf):
        """
        Args:
            surf (EpBunch):
        """
        log(
            'surface "%s" assigned as a Roof' % surf.Name,
            lg.DEBUG,
            name=surf.theidf.name,
        )
        oc = OpaqueConstruction.from_epbunch(
            surf.theidf.getobject("Construction".upper(), surf.Construction_Name)
        )
        oc.area = surf.area
        oc.Surface_Type = "Roof"
        oc.Category = oc.Surface_Type
        return oc

    @staticmethod
    def _do_slab(surf):
        """
        Args:
            surf (EpBunch):
        """
        log(
            'surface "%s" assigned as a Slab' % surf.Name,
            lg.DEBUG,
            name=surf.theidf.name,
        )
        oc = OpaqueConstruction.from_epbunch(
            surf.theidf.getobject("Construction".upper(), surf.Construction_Name)
        )
        oc.area = surf.area
        oc.Surface_Type = "Slab"
        oc.Category = oc.Surface_Type
        return oc

    @staticmethod
    def _do_basement(surf):
        """
        Args:
            surf (EpBunch):
        """
        log(
            'surface "%s" ignored because basement facades are not supported'
            % surf.Name,
            lg.WARNING,
            name=surf.theidf.name,
        )
        oc = None
        return oc

    def mapping(self):
        self.validate()

        return dict(
            Facade=self.Facade,
            Ground=self.Ground,
            Partition=self.Partition,
            Roof=self.Roof,
            Slab=self.Slab,
            IsFacadeAdiabatic=self.IsFacadeAdiabatic,
            IsGroundAdiabatic=self.IsGroundAdiabatic,
            IsPartitionAdiabatic=self.IsPartitionAdiabatic,
            IsRoofAdiabatic=self.IsRoofAdiabatic,
            IsSlabAdiabatic=self.IsSlabAdiabatic,
            Category=self.Category,
            Comments=self.Comments,
            DataSource=self.DataSource,
            Name=self.Name,
        )

    def get_ref(self, ref):
        """Get item matching reference id.

        Args:
            ref:
        """
        return next(
            iter(
                [
                    value
                    for value in ZoneConstructionSet.CREATED_OBJECTS
                    if value.id == ref["$ref"]
                ]
            ),
            None,
        )


def surface_dispatcher(surf, zone):
    """
    Args:
        surf (EpBunch):
        zone (ZoneDefinition):
    """
    dispatch = {
        ("Wall", "Outdoors"): ZoneConstructionSet._do_facade,
        ("Floor", "Ground"): ZoneConstructionSet._do_ground,
        ("Floor", "Outdoors"): ZoneConstructionSet._do_ground,
        ("Floor", "Foundation"): ZoneConstructionSet._do_ground,
        ("Floor", "OtherSideCoefficients"): ZoneConstructionSet._do_ground,
        ("Floor", "GroundSlabPreprocessorAverage"): ZoneConstructionSet._do_ground,
        ("Floor", "Surface"): ZoneConstructionSet._do_slab,
        ("Floor", "Adiabatic"): ZoneConstructionSet._do_slab,
        ("Floor", "Zone"): ZoneConstructionSet._do_slab,
        ("Wall", "Adiabatic"): ZoneConstructionSet._do_partition,
        ("Wall", "Surface"): ZoneConstructionSet._do_partition,
        ("Wall", "Zone"): ZoneConstructionSet._do_partition,
        ("Wall", "Ground"): ZoneConstructionSet._do_basement,
        ("Roof", "Outdoors"): ZoneConstructionSet._do_roof,
        ("Roof", "Zone"): ZoneConstructionSet._do_roof,
        ("Roof", "Surface"): ZoneConstructionSet._do_roof,
        ("Ceiling", "Adiabatic"): ZoneConstructionSet._do_slab,
        ("Ceiling", "Surface"): ZoneConstructionSet._do_slab,
        ("Ceiling", "Zone"): ZoneConstructionSet._do_slab,
    }
    if surf.key.upper() not in ["INTERNALMASS", "WINDOWSHADINGCONTROL"]:
        a, b = surf["Surface_Type"].capitalize(), surf["Outside_Boundary_Condition"]
        try:
            yield dispatch[a, b](surf)
        except KeyError as e:
            raise NotImplementedError(
                "surface '%s' in zone '%s' not supported by surface dispatcher "
                "with keys %s" % (surf.Name, zone.Name, e)
            )



"""EnergyPlus meters module."""

import inspect

import pandas as pd
from geomeppy.patches import EpBunch
from tabulate import tabulate

from archetypal import EnergySeries, ReportData
from archetypal.idfclass.extensions import bunch2db


class Meter:
    """Meter class.

    Holds values for a specific Meter.
    """

    def __init__(self, idf, meter: (dict or EpBunch)):
        """Initialize a Meter object."""
        self._idf = idf
        if isinstance(meter, dict):
            self._key = meter.pop("key").upper()
            self._epobject = self._idf.anidfobject(key=self._key, **meter)
        elif isinstance(meter, EpBunch):
            self._key = meter.key
            self._epobject = meter
        else:
            raise TypeError()

    def __repr__(self):
        """Return the string representation of an EpBunch."""
        return self._epobject.__str__()

    def values(
        self,
        units=None,
        reporting_frequency="Hourly",
        normalize=False,
        sort_values=False,
        ascending=False,
        agg_func="sum",
    ):
        """Return the Meter as a time-series (:class:`EnergySeries`).

        Data is retrieved from the sql file. It is possible to convert the
        time-series to another unit, e.g.: "J" to "kWh".

        Args:
            units (str): Convert original values to another unit. The original unit
                is detected automatically and a dimensionality check is performed.
            reporting_frequency (str): Timestep, Hourly, Daily, Monthly,
                RunPeriod, Environment, Annual or Detailed. Default "Hourly".
            normalize (bool): Normalize between 0 and 1.
            sort_values (bool): If True, values are sorted (default ascending=True)
            ascending (bool): If True and `sort_values` is True, values are sorted in
            ascending order.
            agg_func: #Todo: Document

        Returns:
            EnergySeries: The time-series object.
        """
        self._epobject.Reporting_Frequency = reporting_frequency.lower()
        if self._epobject not in self._idf.idfobjects[self._epobject.key]:
            self._idf.addidfobject(self._epobject)
            self._idf.simulate()
        report = ReportData.from_sqlite(
            sqlite_file=self._idf.sql_file,
            table_name=self._epobject.Key_Name,
            environment_type=1 if self._idf.design_day else 3,
            reporting_frequency=bunch2db[reporting_frequency],
        )
        return EnergySeries.from_reportdata(
            report,
            to_units=units,
            name=self._epobject.Key_Name,
            normalize=normalize,
            sort_values=sort_values,
            ascending=ascending,
            agg_func=agg_func,
        )


class MeterGroup:
    """A class for sub meter groups (Output:Meter vs Output:Meter:Cumulative)."""

    def __init__(self, idf, meters_dict: dict):
        """Initialize MeterGroup."""
        self._idf = idf
        self._properties = {}

        for i, meter in meters_dict.items():
            meter_name = meter["Key_Name"].replace(":", "__").replace(" ", "_")
            self._properties[meter_name] = Meter(idf, meter)
            setattr(self, meter_name, self._properties[meter_name])

    def __getitem__(self, meter_name):
        """Get item by key."""
        return self._properties[meter_name]

    def __repr__(self):
        """Return all available meters."""
        # getmembers() returns all the
        # members of an object
        members = []
        for i in inspect.getmembers(self):

            # to remove private and protected
            # functions
            if not i[0].startswith("_"):

                # To remove other methods that
                # do not start with an underscore
                if not inspect.ismethod(i[1]):
                    members.append(i)

        return f"{len(members)} available meters"


class Meters:
    """Lists available meters in the IDF model.

    Once simulated at least once, the IDF.meters attribute is populated with meters
    categories ("Output:Meter" or "Output:Meter:Cumulative") and each category is
    populated with all the available meters.

    Example:
        For example, to retrieve the WaterSystems:MainsWater meter, simply call

        .. code-block::

            >>> from archetypal import IDF
            >>> idf = IDF()  # load an actual idf file here
            >>> idf.meters.OutputMeter.WaterSystems__MainsWater.values()

    Hint:
        Available meters are read from the .mdd file
    """

    OutputMeter = MeterGroup  # placeholder for OutputMeters

    def __init__(self, idf):
        """Initialize Meter."""
        self._idf = idf

        try:
            mdd, *_ = self._idf.simulation_dir.files("*.mdd")
        except ValueError:
            mdd, *_ = self._idf.simulate().simulation_dir.files("*.mdd")
        if not mdd:
            raise FileNotFoundError
        meters = pd.read_csv(
            mdd, skiprows=2, names=["key", "Key_Name", "Reporting_Frequency"]
        )
        meters.Reporting_Frequency = meters.Reporting_Frequency.str.replace(r"\;.*", "")
        for key, group in meters.groupby("key"):
            meters_dict = group.T.to_dict()
            setattr(
                self,
                key.replace(":", "").replace(" ", "_"),
                MeterGroup(self._idf, meters_dict),
            )

    def __repr__(self):
        """Tabulate available meters."""
        # getmembers() returns all the
        # members of an object
        members = []
        for i in inspect.getmembers(self):

            # to remove private and protected
            # functions
            if not i[0].startswith("_"):

                # To remove other methods that
                # do not start with an underscore
                if not inspect.ismethod(i[1]):
                    members.append(i)
        return tabulate(members, headers=("Available subgroups", "Preview"))



################################################################################
# Module: __init__.py
# Description: Archetypal: Retrieve, construct and analyse building archetypes
# License: MIT, see full license in LICENSE.txt
# Web: https://github.com/samuelduchesne/archetypal
################################################################################

# Version of the package
from setuptools_scm import get_version
__version__ = get_version()

# warn if a newer version of archetypal is available
from outdated import warn_if_outdated

from .eplus_interface.version import warn_if_not_compatible

warn_if_outdated("archetypal", __version__)
warn_if_not_compatible()

# don't display futurewarnings

import warnings

warnings.simplefilter(action="ignore", category=FutureWarning)
warnings.simplefilter(action="ignore", category=UserWarning)

from .utils import *
from .simple_glazing import *
from energy_pandas import EnergySeries, EnergyDataFrame
from .reportdata import ReportData
from .schedule import Schedule
from .plot import *
from .eplus_interface import *
from .idfclass import *
from .dataportal import *
from .umi_template import UmiTemplateLibrary
from .utils import *
from .cli import reduce, transition



import numpy as np
import pytest

import archetypal.settings as settings
from archetypal import IDF
from archetypal.eplus_interface.version import get_eplus_dirs
from archetypal.schedule import Schedule
from archetypal.template.schedule import UmiSchedule, YearSchedule
from archetypal.utils import config


@pytest.fixture()
def schedules_in_necb_specific(config):
    idf = IDF(
        "tests/input_data/necb/NECB 2011-MediumOffice-NECB HDD "
        "Method-CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw.idf"
    )

    s = Schedule(
        Name="NECB-A-Thermostat Setpoint-Heating", idf=idf, start_day_of_the_week=0
    )
    yield s


@pytest.mark.xfail
def test_plot(schedules_in_necb_specific):
    schedules_in_necb_specific.plot(
        slice=("2018/01/02", "2018/01/03"), drawstyle="steps-post"
    )


def test_plot2d(schedules_in_necb_specific):
    schedules_in_necb_specific.plot2d(show=False, save=False)


def test_make_umi_schedule(config):
    """Tests only a single schedule name"""

    idf = IDF("tests/input_data/schedules/schedules.idf")

    s = UmiSchedule(Name="CoolingCoilAvailSched", idf=idf, start_day_of_the_week=0)
    new = s.develop()
    assert hash(s) != hash(new)
    assert id(s) != id(new)

    assert isinstance(s, UmiSchedule)
    assert isinstance(new, YearSchedule)
    assert len(s.all_values) == len(new.all_values)
    np.testing.assert_array_equal(new.all_values, s.all_values)


def test_constant_schedule(config):
    const = Schedule.constant_schedule()
    assert const.__class__.__name__ == "Schedule"


@pytest.fixture()
def mew_idf(config):
    yield IDF(prep_outputs=False)


def test_from_values(mew_idf):
    import numpy as np

    heating_sched = UmiSchedule.from_values(
        Name="Zone_Heating_Schedule",
        Values=np.ones(8760),
        Type="Fraction",
        idf=idf,
    )
    assert len(heating_sched.all_values) == 8760


idf_file = "tests/input_data/schedules/test_multizone_EP.idf"


def schedules_idf():
    config(cache_folder="tests/.temp/cache")
    idf = IDF(
        idf_file,
        epw="tests/input_data/CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw",
        readvars=True,
        include=[
            get_eplus_dirs(settings.ep_version)
            / "DataSets"
            / "TDV"
            / "TDV_2008_kBtu_CTZ06.csv"
        ],
    )
    return idf


idf = schedules_idf()
schedules = list(idf._get_all_schedules(yearly_only=True).keys())
ids = [i.replace(" ", "_") for i in schedules]


@pytest.fixture(scope="module")
def csv_out(config):
    idf = schedules_idf().simulate()
    csv, *_ = idf.simulation_dir.files("*out.csv")
    yield csv


schedules = [
    pytest.param(
        schedule,
        marks=pytest.mark.xfail(
            reason="Can't quite capture all possibilities with special days"
        ),
    )
    if schedule == "POFF"
    else pytest.param(schedule, marks=pytest.mark.xfail(raises=NotImplementedError))
    if schedule == "Cooling Setpoint Schedule"
    else schedule
    for schedule in schedules
]


@pytest.fixture(params=schedules, ids=ids, scope="module")
def schedule_parametrized(request, csv_out):
    """Create the test_data"""
    import pandas as pd

    # read original schedule
    idf = schedules_idf()
    schName = request.param
    orig = Schedule(Name=schName, idf=idf)

    print(
        "{name}\tType:{type}\t[{len}]\tValues:{"
        "values}".format(
            name=orig.Name,
            type=orig.schType,
            values=orig.all_values,
            len=len(orig.all_values),
        )
    )

    # create year:week:day version
    new_eps = orig.to_year_week_day()
    new = orig

    index = orig.series.index
    epv = pd.read_csv(csv_out)
    epv.columns = epv.columns.str.strip()
    epv = epv.loc[:, schName.upper() + ":Schedule Value [](Hourly)"].values
    expected = pd.Series(epv, index=index)

    print("Year: {}".format(new_eps[0].Name))
    print("Weeks: {}".format([obj.Name for obj in new_eps[1]]))
    print("Days: {}".format([obj.Name for obj in new_eps[2]]))

    yield orig, new, expected


def test_ep_versus_schedule(schedule_parametrized):
    """Main test. Will run the idf using EnergyPlus, retrieve the csv file,
    create the schedules and compare"""

    orig, new, expected = schedule_parametrized

    # slice_ = ('2018/01/01 00:00', '2018/01/08 00:00')  # first week
    # slice_ = ('2018/05/20 12:00', '2018/05/22 12:00')
    slice_ = ("2018-10-05 12:00", "2018-10-07 12:00")  # Holiday
    # slice_ = ('2018/01/01 00:00', '2018/12/31 23:00')  # all year
    # slice_ = ('2018/04/30 12:00', '2018/05/01 12:00')  # break

    mask = expected.values.round(3) != orig.all_values.round(3)

    # # region Plot
    # fig, ax = plt.subplots(1, 1, figsize=(5, 4))
    # orig.plot(slice=slice_, ax=ax, legend=True, drawstyle='steps-post',
    #           linestyle='dashed')
    # new.plot(slice=slice_, ax=ax, legend=True, drawstyle='steps-post',
    #          linestyle='dotted')
    # expected.loc[slice_[0]:slice_[1]].plot(label='E+', legend=True, ax=ax,
    #                                        drawstyle='steps-post',
    #                                        linestyle='dashdot')
    # ax.set_title(orig.Name.capitalize())
    # plt.show()
    # # endregion

    print(orig.series[mask])
    assert (
        orig.all_values.round(3)[0 : 52 * 7 * 24] == expected.round(3)[0 : 52 * 7 * 24]
    ).all()
    assert (
        new.all_values.round(3)[0 : 52 * 7 * 24] == expected.round(3)[0 : 52 * 7 * 24]
    ).all()



################################################################################
# Module: archetypal.template
# Description:
# License: MIT, see full license in LICENSE.txt
# Web: https://github.com/samuelduchesne/archetypal
################################################################################

import collections

import numpy as np
from deprecation import deprecated
from sigfig import round

import archetypal
from archetypal.template import MaterialBase


class GasMaterial(MaterialBase):
    """Gas Materials

    .. image:: ../images/template/materials-gas.png
    """

    def __init__(self, Name, Category="Gases", Type="Air", **kwargs):
        """
        Args:
            Name:
            Category:
            Type:
            **kwargs:
        """
        super(GasMaterial, self).__init__(Name, Category=Category, **kwargs)
        self.Type = Type

    def __hash__(self):
        return hash((self.__class__.__name__, getattr(self, "Name", None)))

    def __eq__(self, other):
        if not isinstance(other, GasMaterial):
            return NotImplemented
        else:
            return all(
                [
                    self.Category == other.Category,
                    self.Type == other.Type,
                    self.Conductivity == other.Conductivity,
                    self.Cost == other.Cost,
                    self.Density == other.Density,
                    self.EmbodiedCarbon == other.EmbodiedCarbon,
                    self.EmbodiedEnergy == other.EmbodiedEnergy,
                    np.array_equal(
                        self.SubstitutionRatePattern, other.SubstitutionRatePattern
                    ),
                    self.SubstitutionTimestep == other.SubstitutionTimestep,
                    self.TransportCarbon == other.TransportCarbon,
                    self.TransportDistance == other.TransportDistance,
                    self.TransportEnergy == other.TransportEnergy,
                ]
            )

    @classmethod
    @deprecated(
        deprecated_in="1.3.1",
        removed_in="1.5",
        current_version=archetypal.__version__,
        details="Use from_dict function instead",
    )
    def from_json(cls, *args, **kwargs):

        """
        Args:
            *args:
            **kwargs:
        """
        return cls.from_dict(*args, **kwargs)

    @classmethod
    def from_dict(cls, *args, **kwargs):
        """
        Args:
            *args:
            **kwargs:
        """
        gm = cls(*args, **kwargs)
        gas_type = kwargs.get("Name", None)
        gm.Type = gas_type

        return gm

    def to_json(self):
        """Convert class properties to dict"""
        self.validate()  # Validate object before trying to get json format

        data_dict = collections.OrderedDict()

        data_dict["$id"] = str(self.id)
        data_dict["Category"] = self.Category
        data_dict["Type"] = self.Type
        data_dict["Conductivity"] = round(self.Conductivity, sigfigs=2)
        data_dict["Cost"] = self.Cost
        data_dict["Density"] = self.Density
        data_dict["EmbodiedCarbon"] = self.EmbodiedCarbon
        data_dict["EmbodiedEnergy"] = self.EmbodiedEnergy
        data_dict["SubstitutionRatePattern"] = self.SubstitutionRatePattern
        data_dict["SubstitutionTimestep"] = self.SubstitutionTimestep
        data_dict["TransportCarbon"] = self.TransportCarbon
        data_dict["TransportDistance"] = self.TransportDistance
        data_dict["TransportEnergy"] = self.TransportEnergy
        data_dict["Comments"] = self.Comments
        data_dict["DataSource"] = self.DataSource
        data_dict["Name"] = self.Name

        return data_dict

    def mapping(self):
        self.validate()

        return dict(
            Category=self.Category,
            Type=self.Type,
            Conductivity=self.Conductivity,
            Cost=self.Cost,
            Density=self.Density,
            EmbodiedCarbon=self.EmbodiedCarbon,
            EmbodiedEnergy=self.EmbodiedEnergy,
            SubstitutionRatePattern=self.SubstitutionRatePattern,
            SubstitutionTimestep=self.SubstitutionTimestep,
            TransportCarbon=self.TransportCarbon,
            TransportDistance=self.TransportDistance,
            TransportEnergy=self.TransportEnergy,
            Comments=self.Comments,
            DataSource=self.DataSource,
            Name=self.Name,
        )



"""Window module handles window settings."""

import collections
import logging as lg
from enum import Enum
from functools import reduce

from deprecation import deprecated
from eppy.bunch_subclass import EpBunch

import archetypal
from archetypal import calc_simple_glazing, log, timeit
from archetypal.template import MaterialLayer, UmiSchedule, UniqueName
from archetypal.template.gas_material import GasMaterial
from archetypal.template.glazing_material import GlazingMaterial
from archetypal.template.umi_base import UmiBase


class WindowType(Enum):
    """Refers to the window type. Two choices are available: interior or exterior."""

    External = 0
    Internal = 1

    def __lt__(self, other):
        """Return true if self lower than other."""
        return self._value_ < other._value_

    def __gt__(self, other):
        """Return true if self higher than other."""
        return self._value_ > other._value_


class ShadingType(Enum):
    """Refers to window shading types.

    Hint:
        EnergyPlus specifies 8 different shading types, but only 2 are supported
        here: InteriorShade and ExteriorShade. See shading_ for more info.

    .. _shading: https://bigladdersoftware.com/epx/docs/8-4/input-output-reference/group-thermal-zone-description-geometry.html#field-shading-type
    """

    ExteriorShade = 0
    InteriorShade = 1

    def __lt__(self, other):
        """Return true if self lower than other."""
        return self._value_ < other._value_

    def __gt__(self, other):
        """Return true if self higher than other."""
        return self._value_ > other._value_


class WindowConstruction(UmiBase):
    """Window Construction.

    .. image:: ../images/template/constructions-window.png
    """

    def __init__(
        self,
        Category="Double",
        AssemblyCarbon=0,
        AssemblyCost=0,
        AssemblyEnergy=0,
        DisassemblyCarbon=0,
        DisassemblyEnergy=0,
        Layers=None,
        **kwargs,
    ):
        """Initialize a WindowConstruction.

        Args:
            Category (str): "Single", "Double" or "Triple".
            AssemblyCarbon (float): Assembly Embodied Carbon by m2 of
                construction.
            AssemblyCost (float): Assembly cost by m2 of construction.
            AssemblyEnergy (float): Assembly Embodied Energy by m2; of
                construction.
            DisassemblyCarbon (float): Disassembly embodied carbon by m2 of
                construction.
            DisassemblyEnergy (float): Disassembly embodied energy by m2 of
                construction.
            Layers (list of MaterialLayer):
            **kwargs: Other keywords passed to the constructor.
        """
        super(WindowConstruction, self).__init__(**kwargs)
        self.Category = Category
        self.DisassemblyEnergy = DisassemblyEnergy
        self.DisassemblyCarbon = DisassemblyCarbon
        self.AssemblyEnergy = AssemblyEnergy
        self.AssemblyCost = AssemblyCost
        self.AssemblyCarbon = AssemblyCarbon
        self.Layers = Layers

    def __hash__(self):
        return hash((self.__class__.__name__, getattr(self, "Name", None)))

    def __eq__(self, other):
        if not isinstance(other, WindowConstruction):
            return NotImplemented
        else:
            return all(
                [
                    self.Category == other.Category,
                    self.AssemblyCarbon == other.AssemblyCarbon,
                    self.AssemblyCost == other.AssemblyCost,
                    self.AssemblyEnergy == other.AssemblyEnergy,
                    self.DisassemblyCarbon == other.DisassemblyCarbon,
                    self.DisassemblyEnergy == other.DisassemblyEnergy,
                    self.Layers == other.Layers,
                ]
            )

    @classmethod
    @deprecated(
        deprecated_in="1.3.1",
        removed_in="1.5",
        current_version=archetypal.__version__,
        details="Use from_dict function instead",
    )
    def from_json(cls, *args, **kwargs):

        return cls.from_dict(*args, **kwargs)

    @classmethod
    def from_dict(cls, *args, **kwargs):
        """Create :class:`WindowConstruction` object from json dict."""
        wc = cls(*args, **kwargs)
        layers = kwargs.get("Layers", None)

        # resolve Material objects from ref
        wc.Layers = [
            MaterialLayer(wc.get_ref(layer["Material"]), layer["Thickness"])
            for layer in layers
        ]
        return wc

    @classmethod
    def from_epbunch(cls, Construction, **kwargs):
        """Create :class:`WindowConstruction` object from idf Construction object.

        Example:
            >>> from archetypal import IDF
            >>> from archetypal.template import WindowSetting
            >>> idf = IDF("myidf.idf")
            >>> construction_name = "Some construction name"
            >>> WindowConstruction.from_epbunch(Name=construction_name,
            >>> idf=idf)

        Args:
            Construction (EpBunch): The Construction epbunch object.
            **kwargs: Other keywords passed to the constructor.
        """
        Name = Construction.Name
        idf = Construction.theidf
        wc = cls(Name=Name, idf=idf, **kwargs)
        wc.Layers = wc.layers(Construction, **kwargs)
        catdict = {0: "Single", 1: "Single", 2: "Double", 3: "Triple", 4: "Quadruple"}
        wc.Category = catdict[
            len([lyr for lyr in wc.Layers if isinstance(lyr.Material, GlazingMaterial)])
        ]
        return wc

    def to_json(self):
        """Convert class properties to dict."""
        self.validate()  # Validate object before trying to get json format

        data_dict = collections.OrderedDict()

        data_dict["$id"] = str(self.id)
        data_dict["Layers"] = [layer.to_dict() for layer in self.Layers]
        data_dict["AssemblyCarbon"] = self.AssemblyCarbon
        data_dict["AssemblyCost"] = self.AssemblyCost
        data_dict["AssemblyEnergy"] = self.AssemblyEnergy
        data_dict["DisassemblyCarbon"] = self.DisassemblyCarbon
        data_dict["DisassemblyEnergy"] = self.DisassemblyEnergy
        data_dict["Category"] = self.Category
        data_dict["Comments"] = self.Comments
        data_dict["DataSource"] = self.DataSource
        data_dict["Name"] = UniqueName(self.Name)

        return data_dict

    def mapping(self):
        self.validate()

        return dict(
            Layers=self.Layers,
            AssemblyCarbon=self.AssemblyCarbon,
            AssemblyCost=self.AssemblyCost,
            AssemblyEnergy=self.AssemblyEnergy,
            DisassemblyCarbon=self.DisassemblyCarbon,
            DisassemblyEnergy=self.DisassemblyEnergy,
            Category=self.Category,
            Comments=self.Comments,
            DataSource=self.DataSource,
            Name=self.Name,
        )

    def layers(self, Construction, **kwargs):
        """Retrieve layers for the WindowConstruction"""
        layers = []
        for field in Construction.fieldnames:
            # Loop through the layers from the outside layer towards the
            # indoor layers and get the material they are made of.
            material = Construction.get_referenced_object(field) or kwargs.get(
                "material", None
            )
            if material:
                # Create the WindowMaterial:Glazing or the WindowMaterial:Gas
                # and append to the list of layers
                if material.key.upper() == "WindowMaterial:Glazing".upper():
                    material_obj = GlazingMaterial(
                        Conductivity=material.Conductivity,
                        SolarTransmittance=material.Solar_Transmittance_at_Normal_Incidence,
                        SolarReflectanceFront=material.Front_Side_Solar_Reflectance_at_Normal_Incidence,
                        SolarReflectanceBack=material.Back_Side_Solar_Reflectance_at_Normal_Incidence,
                        VisibleTransmittance=material.Visible_Transmittance_at_Normal_Incidence,
                        VisibleReflectanceFront=material.Front_Side_Visible_Reflectance_at_Normal_Incidence,
                        VisibleReflectanceBack=material.Back_Side_Visible_Reflectance_at_Normal_Incidence,
                        IRTransmittance=material.Infrared_Transmittance_at_Normal_Incidence,
                        IREmissivityFront=material.Front_Side_Infrared_Hemispherical_Emissivity,
                        IREmissivityBack=material.Back_Side_Infrared_Hemispherical_Emissivity,
                        DirtFactor=material.Dirt_Correction_Factor_for_Solar_and_Visible_Transmittance,
                        Type="Uncoated",
                        Name=material.Name,
                        Optical=material.Optical_Data_Type,
                        OpticalData=material.Window_Glass_Spectral_Data_Set_Name,
                        idf=self.idf,
                    )

                    material_layer = MaterialLayer(material_obj, material.Thickness)

                elif material.key.upper() == "WindowMaterial:Gas".upper():
                    # Todo: Make gas name generic, like in UmiTemplateLibrary Editor
                    material_obj = GasMaterial(
                        Name=material.Gas_Type.upper(), idf=self.idf
                    )
                    material_layer = MaterialLayer(material_obj, material.Thickness)
                elif material.key.upper() == "WINDOWMATERIAL:SIMPLEGLAZINGSYSTEM":
                    glass_properties = calc_simple_glazing(
                        material.Solar_Heat_Gain_Coefficient,
                        material.UFactor,
                        material.Visible_Transmittance,
                    )
                    material_obj = GlazingMaterial(
                        **glass_properties, Name=material.Name, idf=self.idf
                    )

                    material_layer = MaterialLayer(
                        material_obj, glass_properties["Thickness"]
                    )
                    layers.append(material_layer)
                    break
                else:
                    continue

                layers.append(material_layer)
        return layers

    def combine(self, other, weights=None):
        """Append other to self. Return self + other as a new object.

        For now, simply returns self.

        todo:
            - Implement equivalent window layers for constant u-factor.

        """
        # Check if other is None. Simply return self
        if not other:
            return self

        if not self:
            return other

        return self

    def validate(self):
        """Validate object and fill in missing values.

        todo:
            - Implement validation
        """
        return self

    def get_ref(self, ref):
        """Get item matching reference id."""
        return next(
            iter(
                [
                    value
                    for value in WindowConstruction.CREATED_OBJECTS
                    if value.id == ref["$ref"]
                ]
            ),
            None,
        )


class WindowSetting(UmiBase):
    """Window Settings define the various window-related properties of a
    specific :class:`Zone`. Control natural ventilation, shading and airflow
    networks and more using this class. This class serves the same role as the
    ZoneInformation>Windows tab in the UMI TemplateEditor.

    .. image:: ../images/template/zoneinfo-windows.png

    Hint:
        The WindowSetting class implements two constructors that are tailored to
        the eppy_ scripting language:

        - :func:`from_construction` and
        - :func:`from_surface`.

    .. _eppy : https://eppy.readthedocs.io/en/latest/
    """

    def __init__(
        self,
        Construction=None,
        OperableArea=0.8,
        AfnWindowAvailability=None,
        AfnDischargeC=0.65,
        AfnTempSetpoint=20,
        IsVirtualPartition=False,
        IsShadingSystemOn=False,
        ShadingSystemAvailabilitySchedule=None,
        ShadingSystemSetpoint=180,
        ShadingSystemTransmittance=0.5,
        ShadingSystemType=ShadingType.ExteriorShade,
        Type=WindowType.External,
        IsZoneMixingOn=False,
        ZoneMixingAvailabilitySchedule=None,
        ZoneMixingDeltaTemperature=2,
        ZoneMixingFlowRate=0.001,
        **kwargs,
    ):
        """Initialize a WindowSetting using default values:

        Args:
            Construction (WindowConstruction): The window construction.
            OperableArea (float): The operable window area as a ratio of total
                window area. eg. 0.8 := 80% of the windows area is operable.
            AfnWindowAvailability (UmiSchedule): The Airflow Network availability
                schedule.
            AfnDischargeC (float): Airflow Network Discharge Coefficient.
                Default = 0.65.
            AfnTempSetpoint (float): Airflow Network Temperature Setpoint.
                Default = 20 degreeC.
            IsVirtualPartition (bool): Virtual Partition.
            IsShadingSystemOn (bool): Shading is used. Default is False.
            ShadingSystemAvailabilitySchedule (UmiSchedule): Shading system
                availability schedule.
            ShadingSystemSetpoint (float): Shading system setpoint in units of
                W/m2. Default = 180 W/m2.
            ShadingSystemTransmittance (float): Shading system transmittance.
                Default = 0.5.
            ShadingSystemType (int): Shading System Type. 0 = ExteriorShade, 1 =
                InteriorShade.
            Type (int):
            IsZoneMixingOn (bool): Zone mixing.
            ZoneMixingAvailabilitySchedule (UmiSchedule): Zone mixing
                availability schedule.
            ZoneMixingDeltaTemperature (float): Zone mixing delta
            ZoneMixingFlowRate (float): Zone mixing flow rate in units of m3/m2.
                Default = 0.001 m3/m2.
            **kwargs: other keywords passed to the constructor.
        """
        super(WindowSetting, self).__init__(**kwargs)

        self.ZoneMixingAvailabilitySchedule = ZoneMixingAvailabilitySchedule
        self.ShadingSystemAvailabilitySchedule = ShadingSystemAvailabilitySchedule
        self.Construction = Construction
        self.AfnWindowAvailability = AfnWindowAvailability
        self.AfnDischargeC = AfnDischargeC
        self.AfnTempSetpoint = AfnTempSetpoint
        self.IsShadingSystemOn = IsShadingSystemOn
        self.IsVirtualPartition = IsVirtualPartition
        self.IsZoneMixingOn = IsZoneMixingOn
        self.OperableArea = OperableArea
        self.ShadingSystemSetpoint = ShadingSystemSetpoint
        self.ShadingSystemTransmittance = ShadingSystemTransmittance
        self.ShadingSystemType = ShadingType(ShadingSystemType)
        self.Type = WindowType(Type)
        self.ZoneMixingDeltaTemperature = ZoneMixingDeltaTemperature
        self.ZoneMixingFlowRate = ZoneMixingFlowRate

    @property
    def OperableArea(self):
        return self._operable_area

    @OperableArea.setter
    def OperableArea(self, value):
        if value > 1:
            raise ValueError("Operable Area must be a number between 0 and 1.")
        self._operable_area = value

    @property
    def AfnDischargeC(self):
        return float(self._afn_discharge_c)

    @AfnDischargeC.setter
    def AfnDischargeC(self, value):
        if value > 1:
            raise ValueError("Operable Area must be a number between 0 and 1.")
        self._afn_discharge_c = value

    @property
    def AfnTempSetpoint(self):
        return float(self._afn_temp_setpoint)

    @AfnTempSetpoint.setter
    def AfnTempSetpoint(self, value):
        self._afn_temp_setpoint = value

    @property
    def ShadingSystemSetpoint(self):
        return float(self._shading_system_setpoint)

    @ShadingSystemSetpoint.setter
    def ShadingSystemSetpoint(self, value):
        self._shading_system_setpoint = value

    @property
    def ShadingSystemTransmittance(self):
        return float(self._shading_system_transmittance)

    @ShadingSystemTransmittance.setter
    def ShadingSystemTransmittance(self, value):
        self._shading_system_transmittance = value

    def __add__(self, other):
        return self.combine(other)

    def __repr__(self):
        # header = "{}: <{}>\n".format(self.Name, self.__class__.mro()[0].__name__)
        # return header + tabulate.tabulate(self.mapping().items(), tablefmt="plain")
        return super(WindowSetting, self).__repr__()

    def __str__(self):
        return repr(self)

    def __hash__(self):
        return hash(
            (self.__class__.__name__, getattr(self, "Name", None), self.DataSource)
        )

    def __eq__(self, other):
        if not isinstance(other, WindowSetting):
            return NotImplemented
        else:
            return all(
                [
                    self.Construction == other.Construction,
                    self.OperableArea == other.OperableArea,
                    self.AfnWindowAvailability == other.AfnWindowAvailability,
                    self.AfnDischargeC == other.AfnDischargeC,
                    self.AfnTempSetpoint == other.AfnTempSetpoint,
                    self.IsVirtualPartition == other.IsVirtualPartition,
                    self.IsShadingSystemOn == other.IsShadingSystemOn,
                    self.ShadingSystemAvailabilitySchedule
                    == other.ShadingSystemAvailabilitySchedule,
                    self.ShadingSystemSetpoint == other.ShadingSystemSetpoint,
                    self.ShadingSystemTransmittance == other.ShadingSystemTransmittance,
                    self.ShadingSystemType == other.ShadingSystemType,
                    self.Type == other.Type,
                    self.IsZoneMixingOn == other.IsZoneMixingOn,
                    self.ZoneMixingAvailabilitySchedule
                    == other.ZoneMixingAvailabilitySchedule,
                    self.ZoneMixingDeltaTemperature == other.ZoneMixingDeltaTemperature,
                    self.ZoneMixingFlowRate == other.ZoneMixingFlowRate,
                ]
            )

    @classmethod
    def generic(cls, idf, Name):
        """Initialize a generic window with SHGC=0.704, UFactor=2.703, Tvis=0.786.

        Args:
            Name (str): Name of the WindowSetting
            idf (IDF):
        """
        material = idf.anidfobject(
            "WindowMaterial:SimpleGlazingSystem".upper(),
            Name="SimpleWindow:SINGLE PANE HW WINDOW",
            UFactor=2.703,
            Solar_Heat_Gain_Coefficient=0.704,
            Visible_Transmittance=0.786,
        )

        constr = idf.anidfobject(
            "CONSTRUCTION",
            Name="SINGLE PANE HW WINDOW",
            Outside_Layer="SimpleWindow:SINGLE PANE HW WINDOW",
        )
        return cls.from_construction(Name=Name, Construction=constr, material=material)

    @classmethod
    def from_construction(cls, Construction, **kwargs):
        """Make a :class:`WindowSetting` directly from a Construction_ object.

        .. _Construction : https://bigladdersoftware.com/epx/docs/8-9/input-output-reference/group-surface-construction-elements.html#construction-000

        Examples:
            >>> from archetypal import IDF
            >>> from archetypal.template import WindowSetting
            >>> # Given an IDF object
            >>> idf = IDF("idfname.idf")
            >>> construction = idf.getobject('CONSTRUCTION',
            >>>                              'AEDG-SmOffice 1A Window Fixed')
            >>> WindowSetting.from_construction(Name='test_window',
            >>>                    Construction=construction)

        Args:
            Construction (EpBunch): The construction name for this window.
            **kwargs: Other keywords passed to the constructor.

        Returns:
            (windowSetting): The window setting object.
        """
        name = kwargs.pop("Name", Construction.Name + "_Window")
        w = cls(Name=name, idf=Construction.theidf, **kwargs)
        w.Construction = WindowConstruction.from_epbunch(Construction, **kwargs)
        w.AfnWindowAvailability = UmiSchedule.constant_schedule(idf=w.idf)
        w.ShadingSystemAvailabilitySchedule = UmiSchedule.constant_schedule(idf=w.idf)
        w.ZoneMixingAvailabilitySchedule = UmiSchedule.constant_schedule(idf=w.idf)
        return w

    @classmethod
    def from_surface(cls, surface, **kwargs):
        """Build a WindowSetting object from a FenestrationSurface:Detailed_.

        This constructor will detect common window constructions and
        shading devices. Supported Shading and Natural Air flow EnergyPlus
        objects are: WindowProperty:ShadingControl_,
        AirflowNetwork:MultiZone:Surface_.

        Important:
            If an EnergyPlus object is not supported, eg.:
            AirflowNetwork:MultiZone:Component:DetailedOpening_, only a warning
            will be issued in the console for the related object instance and
            default values will be automatically used.

        .. _FenestrationSurface:Detailed:
           https://bigladdersoftware.com/epx/docs/8-9/input-output-reference
           /group-thermal-zone-description-geometry.html
           #fenestrationsurfacedetailed
        .. _WindowProperty:ShadingControl:
           https://bigladdersoftware.com/epx/docs/8-9/input-output-reference
           /group-thermal-zone-description-geometry.html
           #windowpropertyshadingcontrol
        .. _AirflowNetwork:MultiZone:Surface:
           https://bigladdersoftware.com/epx/docs/8-9/input-output-reference
           /group-airflow-network.html#airflownetworkmultizonesurface
        .. _AirflowNetwork:MultiZone:Component:DetailedOpening:
           https://bigladdersoftware.com/epx/docs/8-9/input-output-reference
           /group-airflow-network.html
           #airflownetworkmultizonecomponentdetailedopening

        Args:
            surface (EpBunch): The FenestrationSurface:Detailed_ object.

        Returns:
            (WindowSetting): The window setting object.
        """
        if surface.key.upper() == "FENESTRATIONSURFACE:DETAILED":
            construction = surface.get_referenced_object("Construction_Name")
            construction = WindowConstruction.from_epbunch(construction)
            shading_control = surface.get_referenced_object("Shading_Control_Name")
        elif surface.key.upper() == "WINDOW":
            construction = surface.get_referenced_object("Construction_Name")
            construction = WindowConstruction.from_epbunch(construction)
            shading_control = next(
                iter(
                    surface.getreferingobjs(
                        iddgroups=["Thermal Zones and Surfaces"],
                        fields=[f"Fenestration_Surface_{i}_Name" for i in range(1, 10)],
                    )
                ),
                None,
            )
        elif surface.key.upper() == "DOOR":
            return  # Simply skip doors.
        else:
            raise ValueError(
                f"A window of type {surface.key} is not yet supported. "
                f"Please contact developers"
            )

        attr = {}
        if shading_control:
            # a WindowProperty:ShadingControl_ object can be attached to
            # this window
            attr["IsShadingSystemOn"] = True
            if shading_control["Setpoint"] != "":
                attr["ShadingSystemSetpoint"] = shading_control["Setpoint"]
            shade_mat = shading_control.get_referenced_object(
                "Shading_Device_Material_Name"
            )
            # get shading transmittance
            if shade_mat:
                attr["ShadingSystemTransmittance"] = shade_mat["Visible_Transmittance"]
            # get shading control schedule
            if shading_control["Shading_Control_Is_Scheduled"].upper() == "YES":
                name = shading_control["Schedule_Name"]
                attr["ShadingSystemAvailabilitySchedule"] = UmiSchedule(
                    Name=name, idf=surface.theidf
                )
            else:
                # Determine which behavior of control
                shade_ctrl_type = shading_control["Shading_Control_Type"]
                if shade_ctrl_type.lower() == "alwaysoff":
                    attr[
                        "ShadingSystemAvailabilitySchedule"
                    ] = UmiSchedule.constant_schedule(
                        name="AlwaysOff", hourly_value=0, idf=surface.theidf
                    )
                elif shade_ctrl_type.lower() == "alwayson":
                    attr[
                        "ShadingSystemAvailabilitySchedule"
                    ] = UmiSchedule.constant_schedule(idf=surface.theidf)
                else:
                    log(
                        'Window "{}" uses a  window control type that '
                        'is not supported: "{}". Reverting to '
                        '"AlwaysOn"'.format(surface.Name, shade_ctrl_type),
                        lg.WARN,
                    )
                    attr[
                        "ShadingSystemAvailabilitySchedule"
                    ] = UmiSchedule.constant_schedule(idf=surface.theidf)
            # get shading type
            if shading_control["Shading_Type"] != "":
                mapping = {
                    "InteriorShade": ShadingType(1),
                    "ExteriorShade": ShadingType(0),
                    "ExteriorScreen": ShadingType(0),
                    "InteriorBlind": ShadingType(1),
                    "ExteriorBlind": ShadingType(0),
                    "BetweenGlassShade": ShadingType(0),
                    "BetweenGlassBlind": ShadingType(0),
                    "SwitchableGlazing": ShadingType(0),
                }
                attr["ShadingSystemType"] = mapping[shading_control["Shading_Type"]]
        else:
            # Set default schedules
            attr["ShadingSystemAvailabilitySchedule"] = UmiSchedule.constant_schedule(
                idf=surface.theidf
            )

        # get airflow network
        afn = next(
            iter(
                surface.getreferingobjs(
                    iddgroups=["Natural Ventilation and Duct Leakage"],
                    fields=["Surface_Name"],
                )
            ),
            None,
        )

        if afn:
            attr["OperableArea"] = afn.WindowDoor_Opening_Factor_or_Crack_Factor
            leak = afn.get_referenced_object("Leakage_Component_Name")
            name = afn["Venting_Availability_Schedule_Name"]
            if name != "":
                attr["AfnWindowAvailability"] = UmiSchedule(
                    Name=name, idf=surface.theidf
                )
            else:
                attr["AfnWindowAvailability"] = UmiSchedule.constant_schedule(
                    idf=surface.theidf
                )
            name = afn["Ventilation_Control_Zone_Temperature_Setpoint_Schedule_Name"]
            if name != "":
                attr["AfnTempSetpoint"] = UmiSchedule(
                    Name=name, idf=surface.theidf
                ).mean
            else:
                pass  # uses default

            if (
                leak.key.upper()
                == "AIRFLOWNETWORK:MULTIZONE:SURFACE:EFFECTIVELEAKAGEAREA"
            ):
                attr["AfnDischargeC"] = leak["Discharge_Coefficient"]
            elif (
                leak.key.upper()
                == "AIRFLOWNETWORK:MULTIZONE:COMPONENT:HORIZONTALOPENING"
            ):
                log(
                    '"{}" is not fully supported. Reverting to '
                    'defaults for object "{}"'.format(leak.key, cls.mro()[0].__name__),
                    lg.WARNING,
                )
            elif leak.key.upper() == "AIRFLOWNETWORK:MULTIZONE:SURFACE:CRACK":
                log(
                    '"{}" is not fully supported. Rerverting to '
                    'defaults for object "{}"'.format(leak.key, cls.mro()[0].__name__),
                    lg.WARNING,
                )
            elif (
                leak.key.upper() == "AIRFLOWNETWORK:MULTIZONE:COMPONENT:DETAILEDOPENING"
            ):
                log(
                    '"{}" is not fully supported. Rerverting to '
                    'defaults for object "{}"'.format(leak.key, cls.mro()[0].__name__),
                    lg.WARNING,
                )
            elif (
                leak.key.upper() == "AIRFLOWNETWORK:MULTIZONE:COMPONENT:ZONEEXHAUSTFAN"
            ):
                log(
                    '"{}" is not fully supported. Rerverting to '
                    'defaults for object "{}"'.format(leak.key, cls.mro()[0].__name__),
                    lg.WARNING,
                )
            elif leak.key.upper() == "AIRFLOWNETWORK:MULTIZONE:COMPONENT:SIMPLEOPENING":
                log(
                    '"{}" is not fully supported. Rerverting to '
                    'defaults for object "{}"'.format(leak.key, cls.mro()[0].__name__),
                    lg.WARNING,
                )
        else:
            attr["AfnWindowAvailability"] = UmiSchedule.constant_schedule(
                hourly_value=0, Name="AlwaysOff", idf=surface.theidf
            )
        # Todo: Zone Mixing is always off
        attr["ZoneMixingAvailabilitySchedule"] = UmiSchedule.constant_schedule(
            hourly_value=0, Name="AlwaysOff", idf=surface.theidf
        )
        DataSource = kwargs.pop("DataSource", surface.theidf.name)
        Category = kwargs.pop("Category", surface.theidf.name)
        w = cls(
            Name=surface.Name,
            Construction=construction,
            idf=surface.theidf,
            Category=Category,
            DataSource=DataSource,
            **attr,
            **kwargs,
        )
        return w

    @classmethod
    @timeit
    def from_zone(cls, zone, **kwargs):
        """Iterate over the zone subsurfaces and create a window object.

        If more than one window is created, use reduce to combine them together.

        Args:
            zone (Zone): The Zone object from which the WindowSetting is
                created.

        Returns:
            WindowSetting: The WindowSetting object for this zone.
        """
        window_sets = []

        for surf in zone._zonesurfaces:
            # skip internalmass objects since they don't have windows.
            if surf.key.lower() != "internalmass":
                for subsurf in surf.subsurfaces:
                    # For each subsurface, create a WindowSetting object
                    # using the `from_surface` constructor.
                    window_sets.append(cls.from_surface(subsurf, **kwargs))

        if window_sets:
            # if one or more window has been created, reduce. Using reduce on
            # a len==1 list, will simply return the object.

            return reduce(WindowSetting.combine, window_sets)
        else:
            # no window found, probably a core zone, return None.
            return None

    def combine(self, other, weights=None, allow_duplicates=False):
        """Append other to self. Return self + other as a new object.

        Args:
            other (WindowSetting): The other OpaqueMaterial object
            weights (list-like, optional): A list-like object of len 2. If None,
                equal weights are used.

        Returns:
            WindowSetting: A new combined object made of self + other.
        """
        # Check if other is None. Simply return self
        if not other:
            return self

        if not self:
            return other

        if not isinstance(other, self.__class__):
            msg = "Cannot combine %s with %s" % (
                self.__class__.__name__,
                other.__class__.__name__,
            )
            raise NotImplementedError(msg)

        # Check if other is not the same as self
        if self == other:
            return self

        if not weights:
            log(
                'using 1 as weighting factor in "{}" '
                "combine.".format(self.__class__.__name__)
            )
            weights = [1.0, 1.0]
        meta = self._get_predecessors_meta(other)
        new_attr = dict(
            Construction=WindowConstruction.combine(
                self.Construction, other.Construction, weights
            ),
            AfnDischargeC=self._float_mean(other, "AfnDischargeC", weights),
            AfnTempSetpoint=self._float_mean(other, "AfnTempSetpoint", weights),
            AfnWindowAvailability=UmiSchedule.combine(
                self.AfnWindowAvailability, other.AfnWindowAvailability, weights
            ),
            IsShadingSystemOn=any([self.IsShadingSystemOn, other.IsShadingSystemOn]),
            IsVirtualPartition=any([self.IsVirtualPartition, other.IsVirtualPartition]),
            IsZoneMixingOn=any([self.IsZoneMixingOn, other.IsZoneMixingOn]),
            OperableArea=self._float_mean(other, "OperableArea", weights),
            ShadingSystemSetpoint=self._float_mean(
                other, "ShadingSystemSetpoint", weights
            ),
            ShadingSystemTransmittance=self._float_mean(
                other, "ShadingSystemTransmittance", weights
            ),
            ShadingSystemType=max(self.ShadingSystemType, other.ShadingSystemType),
            ZoneMixingDeltaTemperature=self._float_mean(
                other, "ZoneMixingDeltaTemperature", weights
            ),
            ZoneMixingFlowRate=self._float_mean(other, "ZoneMixingFlowRate", weights),
            ZoneMixingAvailabilitySchedule=UmiSchedule.combine(
                self.ZoneMixingAvailabilitySchedule,
                other.ZoneMixingAvailabilitySchedule,
                weights,
            ),
            ShadingSystemAvailabilitySchedule=UmiSchedule.combine(
                self.ShadingSystemAvailabilitySchedule,
                other.ShadingSystemAvailabilitySchedule,
                weights,
            ),
            Type=max(self.Type, other.Type),
        )
        new_obj = WindowSetting(**meta, **new_attr, idf=self.idf)
        new_obj.predecessors.update(self.predecessors + other.predecessors)
        return new_obj

    def to_json(self):
        """Convert class properties to dict."""
        self.validate()  # Validate object before trying to get json format

        data_dict = collections.OrderedDict()

        data_dict["$id"] = str(self.id)
        data_dict["AfnDischargeC"] = self.AfnDischargeC
        data_dict["AfnTempSetpoint"] = self.AfnTempSetpoint
        data_dict["AfnWindowAvailability"] = self.AfnWindowAvailability.to_dict()
        data_dict["Construction"] = {"$ref": str(self.Construction.id)}
        data_dict["IsShadingSystemOn"] = self.IsShadingSystemOn
        data_dict["IsVirtualPartition"] = self.IsVirtualPartition
        data_dict["IsZoneMixingOn"] = self.IsZoneMixingOn
        data_dict["OperableArea"] = self.OperableArea
        data_dict[
            "ShadingSystemAvailabilitySchedule"
        ] = self.ShadingSystemAvailabilitySchedule.to_dict()
        data_dict["ShadingSystemSetpoint"] = self.ShadingSystemSetpoint
        data_dict["ShadingSystemTransmittance"] = self.ShadingSystemTransmittance
        data_dict["ShadingSystemType"] = self.ShadingSystemType.value
        data_dict["Type"] = self.Type.value
        data_dict[
            "ZoneMixingAvailabilitySchedule"
        ] = self.ZoneMixingAvailabilitySchedule.to_dict()
        data_dict["ZoneMixingDeltaTemperature"] = self.ZoneMixingDeltaTemperature
        data_dict["ZoneMixingFlowRate"] = self.ZoneMixingFlowRate
        data_dict["Category"] = self.Category
        data_dict["Comments"] = self.Comments
        data_dict["DataSource"] = self.DataSource
        data_dict["Name"] = UniqueName(self.Name)

        return data_dict

    @classmethod
    @deprecated(
        deprecated_in="1.3.1",
        removed_in="1.5",
        current_version=archetypal.__version__,
        details="Use from_dict function instead",
    )
    def from_json(cls, *args, **kwargs):

        return cls.from_dict(*args, **kwargs)

    @classmethod
    def from_dict(cls, *args, **kwargs):
        """Initialize :class:`WindowSetting` object from json dict."""
        w = cls(*args, **kwargs)

        ref = kwargs.get("AfnWindowAvailability", None)
        w.AfnWindowAvailability = w.get_ref(ref)
        ref = kwargs.get("Construction", None)
        w.Construction = w.get_ref(ref)
        ref = kwargs.get("ShadingSystemAvailabilitySchedule", None)
        w.ShadingSystemAvailabilitySchedule = w.get_ref(ref)
        ref = kwargs.get("ZoneMixingAvailabilitySchedule", None)
        w.ZoneMixingAvailabilitySchedule = w.get_ref(ref)
        return w

    @classmethod
    def from_ref(cls, ref, building_templates, idf=None, **kwargs):
        """Initialize :class:`WindowSetting` object from a reference id.

        Hint:
            In some cases, the WindowSetting is referenced in the DataStore to the
            Windows property of a BuildingTemplate (instead of being listed in the
            WindowSettings list. This is the case in the original
            BostonTemplateLibrary.json.

        Args:
            ref (str): The referenced number in the json library.
            building_templates (list): List of BuildingTemplates from the datastore.

        Returns:
            WindowSetting: The parsed WindowSetting.
        """
        store = next(
            iter(
                filter(
                    lambda x: x.get("$id") == ref,
                    [bldg.get("Windows") for bldg in building_templates],
                )
            )
        )
        w = cls.from_json(**store, idf=idf, **kwargs)
        return w

    def validate(self):
        """Validate object and fill in missing values."""
        if not self.AfnWindowAvailability:
            self.AfnWindowAvailability = UmiSchedule.constant_schedule(
                hourly_value=0, Name="AlwaysOff", idf=self.idf
            )
        if not self.ShadingSystemAvailabilitySchedule:
            self.ShadingSystemAvailabilitySchedule = UmiSchedule.constant_schedule(
                hourly_value=0, Name="AlwaysOff", idf=self.idf
            )
        if not self.ZoneMixingAvailabilitySchedule:
            self.ZoneMixingAvailabilitySchedule = UmiSchedule.constant_schedule(
                hourly_value=0, Name="AlwaysOff", idf=self.idf
            )

        return self

    def mapping(self):
        self.validate()

        return dict(
            AfnDischargeC=self.AfnDischargeC,
            AfnTempSetpoint=self.AfnTempSetpoint,
            AfnWindowAvailability=self.AfnWindowAvailability,
            Construction=self.Construction,
            IsShadingSystemOn=self.IsShadingSystemOn,
            IsVirtualPartition=self.IsVirtualPartition,
            IsZoneMixingOn=self.IsZoneMixingOn,
            OperableArea=self.OperableArea,
            ShadingSystemAvailabilitySchedule=self.ShadingSystemAvailabilitySchedule,
            ShadingSystemSetpoint=self.ShadingSystemSetpoint,
            ShadingSystemTransmittance=self.ShadingSystemTransmittance,
            ShadingSystemType=self.ShadingSystemType,
            Type=self.Type,
            ZoneMixingAvailabilitySchedule=self.ZoneMixingAvailabilitySchedule,
            ZoneMixingDeltaTemperature=self.ZoneMixingDeltaTemperature,
            ZoneMixingFlowRate=self.ZoneMixingFlowRate,
            Category=self.Category,
            Comments=self.Comments,
            DataSource=self.DataSource,
            Name=self.Name,
        )

    def get_ref(self, ref):
        """Get item matching reference id.

        Args:
            ref:
        """
        return next(
            iter(
                [
                    value
                    for value in WindowSetting.CREATED_OBJECTS
                    if value.id == ref["$ref"]
                ]
            ),
            None,
        )



################################################################################
# Module: settings.py
# Description: Various settings used across the package
# License: MIT, see full license in LICENSE.txt
# Web: https://github.com/samuelduchesne/archetypal
################################################################################

import logging as lg
import os

import pint
from path import Path

# locations to save data, logs, images, and cache
import archetypal

data_folder = Path("data")
logs_folder = Path("logs")
imgs_folder = Path("images")
cache_folder = Path("cache")
umitemplate = Path("data/BostonTemplateLibrary.json")

# cache server responses
use_cache = False

# Debug behavior
debug = False

# write log to file and/or to console
log_file = False
log_console = False
log_notebook = False
log_level = lg.INFO
log_name = "archetypal"
log_filename = "archetypal"

# usual idfobjects
useful_idf_objects = [
    "WINDOWMATERIAL:GAS",
    "WINDOWMATERIAL:GLAZING",
    "WINDOWMATERIAL:SIMPLEGLAZINGSYSTEM",
    "MATERIAL",
    "MATERIAL:NOMASS",
    "CONSTRUCTION",
    "BUILDINGSURFACE:DETAILED",
    "FENESTRATIONSURFACE:DETAILED",
    "SCHEDULE:DAY:INTERVAL",
    "SCHEDULE:WEEK:DAILY",
    "SCHEDULE:YEAR",
]

# List of Available SQLite Tables
# Ref: https://bigladdersoftware.com/epx/docs/8-3/output-details-and-examples
# /eplusout.sql.html#schedules-table

available_sqlite_tables = dict(
    ComponentSizes={"PrimaryKey": ["ComponentSizesIndex"], "ParseDates": []},
    ConstructionLayers={"PrimaryKey": ["ConstructionIndex"], "ParseDates": []},
    Constructions={"PrimaryKey": ["ConstructionIndex"], "ParseDates": []},
    Materials={"PrimaryKey": ["MaterialIndex"], "ParseDates": []},
    NominalBaseboardHeaters={
        "PrimaryKey": ["NominalBaseboardHeaterIndex"],
        "ParseDates": [],
    },
    NominalElectricEquipment={
        "PrimaryKey": ["NominalElectricEquipmentIndex"],
        "ParseDates": [],
    },
    NominalGasEquipment={"PrimaryKey": ["NominalGasEquipmentIndex"], "ParseDates": []},
    NominalHotWaterEquipment={
        "PrimaryKey": ["NominalHotWaterEquipmentIndex"],
        "ParseDates": [],
    },
    NominalInfiltration={"PrimaryKey": ["NominalInfiltrationIndex"], "ParseDates": []},
    NominalLighting={"PrimaryKey": ["NominalLightingIndex"], "ParseDates": []},
    NominalOtherEquipment={
        "PrimaryKey": ["NominalOtherEquipmentIndex"],
        "ParseDates": [],
    },
    NominalPeople={"PrimaryKey": ["NominalPeopleIndex"], "ParseDates": []},
    NominalSteamEquipment={
        "PrimaryKey": ["NominalSteamEquipmentIndex"],
        "ParseDates": [],
    },
    NominalVentilation={"PrimaryKey": ["NominalVentilationIndex"], "ParseDates": []},
    ReportData={"PrimaryKey": ["ReportDataIndex"], "ParseDates": []},
    ReportDataDictionary={
        "PrimaryKey": ["ReportDataDictionaryIndex"],
        "ParseDates": [],
    },
    ReportExtendedData={"PrimaryKey": ["ReportExtendedDataIndex"], "ParseDates": []},
    RoomAirModels={"PrimaryKey": ["ZoneIndex"], "ParseDates": []},
    Schedules={"PrimaryKey": ["ScheduleIndex"], "ParseDates": []},
    Surfaces={"PrimaryKey": ["SurfaceIndex"], "ParseDates": []},
    SystemSizes={
        "PrimaryKey": ["SystemSizesIndex"],
        "ParseDates": {"PeakHrMin": "%m/%d %H:%M:%S"},
    },
    Time={"PrimaryKey": ["TimeIndex"], "ParseDates": []},
    ZoneGroups={"PrimaryKey": ["ZoneGroupIndex"], "ParseDates": []},
    Zones={"PrimaryKey": ["ZoneIndex"], "ParseDates": []},
    ZoneLists={"PrimaryKey": ["ZoneListIndex"], "ParseDates": []},
    ZoneSizes={"PrimaryKey": ["ZoneSizesIndex"], "ParseDates": []},
    ZoneInfoZoneLists={"PrimaryKey": ["ZoneListIndex"], "ParseDates": []},
    Simulations={
        "PrimaryKey": ["SimulationIndex"],
        "ParseDates": {"TimeStamp": {"format": "YMD=%Y.%m.%d %H:%M"}},
    },
    EnvironmentPeriods={"PrimaryKey": ["EnvironmentPeriodIndex"], "ParseDates": []},
    TabularData={"PrimaryKey": ["TabularDataIndex"], "ParseDates": []},
    Strings={"PrimaryKey": ["StringIndex"], "ParseDates": []},
    StringTypes={"PrimaryKey": ["StringTypeIndex"], "ParseDates": []},
    TabularDataWithStrings={"PrimaryKey": ["TabularDataIndex"], "ParseDates": []},
    Errors={"PrimaryKey": ["ErrorIndex"], "ParseDates": []},
)

# common_umi_objects
common_umi_objects = []

# default crs for when creating plots and querying databases
default_crs = {"init": "epsg:4326"}

# unique schedule number as list
unique_schedules = []

# region read template - use io.BytesIO(settings.template) in code
import pkg_resources

resource_package = archetypal.__name__  # Could be any module/package name

# originBUISketchUp.idf template
resource_path = "/".join(("ressources", "originBUISketchUp.idf"))
# Do not use os.path.join()
template_BUI = pkg_resources.resource_string(resource_package, resource_path)

# window library ('W74-lib.dat') template
resource_path = "/".join(("ressources", "W74-lib.dat"))
# Do not use os.path.join()
template_winLib = pkg_resources.resource_string(resource_package, resource_path)

# NewFileTemplate.d18 ('W74-lib.dat') template
resource_path = "/".join(("ressources", "NewFileTemplate.d18"))
# Do not use os.path.join()
path_template_d18 = pkg_resources.resource_filename(resource_package, resource_path)
# endregion

# Units

from energy_pandas.units import unit_registry

unit_registry.define("m3 = 1 * meter ** 3 = m³")
unit_registry.define(
    "degree_Celsius = kelvin; offset: 273.15 = °C = C = celsius = degC = degreeC"
)
unit_registry.define(
    "degree_Fahrenheit = 5 / 9 * kelvin; offset: 233.15 + 200 / 9 = "
    "°F = F = fahrenheit = degF = degreeF"
)
unit_registry.define("ach = dimensionless")  # Air Changes per Hour
unit_registry.define("acr = 1 / hour")  # Air change rate


class ZoneWeight(object):
    """Zone weights for Umi Templates"""

    weight_attr = {0: "area", 1: "volume"}

    def __init__(self, n=0):
        self._weight_attr = self.weight_attr[n]

    def __str__(self):
        return self.get_weight_attr()

    def get_weight_attr(self):
        return self._weight_attr

    def set_weigth_attr(self, weight):
        if weight not in self.weight_attr.values():
            i = len(self.weight_attr) + 1
            self.weight_attr[i] = weight
        self._weight_attr = weight


zone_weight = ZoneWeight(n=0)

# Latest version of EnergyPlus compatible with archetypal
ep_version = "9-2-0"



"""Transition module."""

import os
import platform
import re
import shutil
import subprocess
import time
from io import StringIO
from subprocess import CalledProcessError
from threading import Thread
import logging as lg

from eppy.runner.run_functions import paths_from_version
from path import Path
from tqdm import tqdm

from archetypal.eplus_interface.energy_plus import EnergyPlusProgram
from archetypal.eplus_interface.exceptions import (
    EnergyPlusProcessError,
    EnergyPlusVersionError,
)
from archetypal.eplus_interface.version import EnergyPlusVersion
from archetypal.utils import log


class TransitionExe(EnergyPlusProgram):
    """Transition Program Generator.

    Examples:
        >>> from archetypal import IDF
        >>> for transition in TransitionExe(IDF(), tmp_dir=os.getcwd()):
        >>>     print(transition.cmd())
    """

    def __init__(self, idf, tmp_dir):
        """Initialize Transition Executable."""
        super().__init__(idf)
        self.idf = idf
        self.trans = None  # Set by __next__()
        self.running_directory = tmp_dir

        self._trans_exec = None

    def __next__(self):
        """Return next transition."""
        self.trans = next(self.transitions_generator)
        return self

    def __iter__(self):
        """Iterate over transitions."""
        return self

    def get_exe_path(self):
        """Return the path containing the next transition."""
        if not self.trans_exec[self.trans].exists():
            raise EnergyPlusProcessError(
                cmd=self.trans_exec[self.trans],
                stderr="The specified EnergyPlus version (v{}) does not have"
                " the required transition program '{}' in the "
                "PreProcess folder. See the documentation "
                "(archetypal.readthedocs.io/troubleshooting.html#missing"
                "-transition-programs) "
                "to solve this issue".format(
                    self.idf.as_version, self.trans_exec[self.trans]
                ),
                idf=self.idf,
            )
        return self.trans_exec[self.trans]

    @property
    def idfname(self):
        """Copy and return self.idf to the output directory and expand."""
        return Path(self.idf.savecopy(self.running_directory / "in.idf")).expand()

    @property
    def trans_exec(self) -> dict:
        """Return dict of {EnergyPlusVersion: executable} for each transitions."""

        def copytree(src, dst, symlinks=False, ignore=None):
            for item in os.listdir(src):
                s = os.path.join(src, item)
                d = os.path.join(dst, item)
                try:
                    if os.path.isdir(s):
                        shutil.copytree(s, d, symlinks, ignore)
                    else:
                        shutil.copy2(s, d)
                except FileNotFoundError as e:
                    time.sleep(60)
                    log(f"{e}")

        if self._trans_exec is None:
            copytree(self.idf.idfversionupdater_dir, self.running_directory)
            self._trans_exec = {
                EnergyPlusVersion(
                    re.search(r"to-V(([\d])-([\d])-([\d]))", exec).group(1)
                ): exec
                for exec in self.running_directory.files("Transition-V*")
            }
        return self._trans_exec

    @property
    def transitions(self) -> list:
        """Return a sorted list of necessary transitions."""
        transitions = [
            key
            for key in self.trans_exec
            if self.idf.as_version >= key > self.idf.file_version
        ]
        transitions.sort()
        return transitions

    @property
    def transitions_generator(self):
        """Generate transitions."""
        for transition in self.transitions:
            yield transition

    def __str__(self):
        """Return string representation."""
        return " ".join(self.__repr__())

    def __repr__(self):
        """Return command as string."""
        return self.cmd()

    def cmd(self):
        """Get the platform-specific command."""
        _which = Path(shutil.which(self.get_exe_path()))
        if platform.system() == "Windows":
            cmd = [_which.relpath(), self.idfname.basename()]
        else:
            # must specify current dir on Unix
            cmd = ["./" + _which.basename(), self.idfname.basename()]
        return cmd


class TransitionThread(Thread):
    """Transition program manager."""

    def __init__(self, idf, tmp, overwrite=False):
        """Initialize Thread."""
        super(TransitionThread, self).__init__()
        self.overwrite = overwrite
        self.p = None
        self.std_out = None
        self.std_err = None
        self.idf = idf
        self.cancelled = False
        self.run_dir = Path("")
        self.exception = None
        self.name = "Transition_" + self.idf.name
        self.tmp = tmp
        self.idfname = None
        self.idd = None
        self.cmd = None

    def run(self):
        """Run.

        Wrapper around the EnergyPlus command line interface.
        """
        self.cancelled = False
        # get version from IDF object or by parsing the IDF file for it

        # Move files into place
        tmp = self.tmp
        self.idfname = Path(self.idf.savecopy(tmp / "in.idf")).expand()
        self.idd = self.idf.iddname.copy(tmp).expand()

        generator = TransitionExe(self.idf, tmp_dir=tmp)

        for trans in tqdm(
            generator,
            total=len(generator.transitions),
            unit_scale=True,
            miniters=1,
            position=self.idf.position,
            desc=f"Transition #{self.idf.position}-{self.idf.name}",
        ):
            # Get executable using shutil.which (determines the extension
            # based on the platform, eg: .exe. And copy the executable to tmp
            self.run_dir = Path(tmp).expand()

            # Run Transition Program
            self.cmd = trans.cmd()
            self.p = subprocess.Popen(
                self.cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                shell=False,  # cannot use shell
                cwd=self.run_dir,
            )
            start_time = time.time()
            self.msg_callback("Transition started")
            for line in self.p.stdout:
                self.msg_callback(line.decode("utf-8").strip("\n"))

            # We explicitly close stdout
            self.p.stdout.close()

            # Wait for process to complete
            self.p.wait()

            # Communicate callbacks
            if self.cancelled:
                self.msg_callback("Transition cancelled")
                # self.cancelled_callback(self.std_out, self.std_err)
            else:
                if self.p.returncode == 0:
                    self.msg_callback(
                        "Transition completed in {:,.2f} seconds".format(
                            time.time() - start_time
                        )
                    )
                    self.success_callback()
                    for line in self.p.stderr:
                        self.msg_callback(line.decode("utf-8"))
                else:
                    self.msg_callback("Transition failed")
                    self.failure_callback()

    @property
    def trans_exec(self) -> dict:
        """Return dict of {EnergyPlusVersion, executable} for each transitions."""
        return {
            EnergyPlusVersion(
                re.search(r"to-V(([\d])-([\d])-([\d]))", exec).group(1)
            ): exec
            for exec in self.idf.idfversionupdater_dir.files("Transition-V*")
        }

    @property
    def transitions(self):
        """Return a sorted list of necessary transitions."""
        transitions = [
            key
            for key in self.trans_exec
            if self.idf.as_version >= key > self.idf.file_version
        ]
        transitions.sort()
        return transitions

    def msg_callback(self, *args, **kwargs):
        """Pass message to logger."""
        log(*args, name=self.idf.name, **kwargs)

    def success_callback(self):
        """Retrieve the transitioned file.

        If self.overwrite is True, the transitioned file replaces the
        original file.
        """
        for f in Path(self.run_dir).files("*.idfnew"):
            if isinstance(self.idf.idfname, StringIO) or not self.overwrite:
                file = StringIO(f.read_text())
            else:
                file = f.copy(self.idf.idfname)

            # replace idfname with file
            try:
                self.idf.idfname = file
            except (NameError, UnboundLocalError):
                self.exception = EnergyPlusProcessError(
                    cmd="IDF.upgrade",
                    stderr="An error occurred during transitioning",
                    idf=self.idf,
                )
            else:
                self.idf._reset_dependant_vars("idfname")

    def failure_callback(self):
        """Read stderr and pass to logger."""
        for line in self.p.stderr:
            self.msg_callback(line.decode("utf-8"), level=lg.ERROR)
        raise CalledProcessError(self.p.returncode, cmd=self.cmd, stderr=self.p.stderr)

    def cancelled_callback(self, stdin, stdout):
        """Call on cancelled."""
        pass

    @property
    def eplus_home(self):
        """Return the location of the EnergyPlus directory."""
        eplus_exe, eplus_home = paths_from_version(self.idf.as_version.dash)
        if not Path(eplus_home).exists():
            raise EnergyPlusVersionError(
                msg=f"No EnergyPlus Executable found for version "
                f"{EnergyPlusVersion(self.idf.as_version)}"
            )
        else:
            return Path(eplus_home)



import pytest
from numpy.testing import assert_almost_equal
from pandas import read_csv

from archetypal import IDF, EnergyDataFrame, EnergySeries, settings


@pytest.fixture(scope="module")
def idf(config):
    idfname = "tests/input_data/umi_samples/B_Off_0.idf"
    epw = "tests/input_data/CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw"
    _idf = IDF(idfname, epw)
    yield _idf
    _idf.save()


@pytest.fixture(scope="class")
def rd_es(idf):
    try:
        yield idf.meters.OutputMeter.WaterSystems__MainsWater.values()
    except Exception:
        yield idf.simulate().meters.OutputMeter.WaterSystems__MainsWater.values()


@pytest.fixture(scope="class")
def rd_edf(idf):
    try:
        yield idf.variables.OutputVariable.Zone_Air_Temperature.values()
    except Exception:
        yield idf.simulate().variables.OutputVariable.Zone_Air_Temperature.values()


class TestEnergySeries:
    def test_from_csv(self):
        file = "tests/input_data/test_profile.csv"
        df = read_csv(file, index_col=[0], names=["Heat"])
        ep = EnergySeries.with_timeindex(df.Heat, frequency="1H", units="BTU/hour")
        assert ep.units == settings.unit_registry.parse_expression("BTU/hour").units

    def test_from_report_data(self, rd_es):
        assert_almost_equal(rd_es.sum(), 2.1187133811706036, decimal=3)
        assert rd_es.units == settings.unit_registry.m3

    @pytest.mark.parametrize("kind", ["polygon", "surface", "contour"])
    def test_plot_3d(self, rd_es, kind):
        fig, ax = rd_es.plot3d(
            save=False,
            show=True,
            axis_off=False,
            kind=kind,
            cmap="Reds",
            fig_width=4,
            fig_height=4,
            edgecolors="grey",
            linewidths=0.01,
        )

    def test_plot_2d(self, rd_es):
        fig, ax = rd_es.plot2d(
            axis_off=False,
            cmap="Reds",
            fig_height=2,
            fig_width=6,
            show=True,
            save=False,
            filename=rd_es.name + "_heatmap",
        )

    def test_discretize(self, rd_es):
        res = rd_es.discretize_tsam(noTypicalPeriods=1)
        assert_almost_equal(res.sum(), 2.118713381170598, decimal=3)
        rd_es.discretize_tsam(noTypicalPeriods=1, inplace=True)
        assert_almost_equal(res.sum(), 2.118713381170598, decimal=3)
        # check that the type is maintained
        assert type(rd_es) == EnergySeries


class TestEnergyDataFrame:
    def test_from_report_data(self, rd_edf):
        assert_almost_equal(rd_edf.sum().sum(), 397714.38260559476, decimal=3)
        assert rd_edf.units["CORE"] == settings.unit_registry.C

    def test_discretize(self, rd_edf):
        rd_edf = rd_edf.discretize_tsam(noTypicalPeriods=1)
        assert hasattr(rd_edf, "agg")
        rd_edf.discretize_tsam(noTypicalPeriods=1, inplace=True)
        assert hasattr(rd_edf, "agg")
        # check that the type is maintained
        assert type(rd_edf) == EnergyDataFrame

    def test_plot_2d(self, rd_edf):
        fig, ax = rd_edf.plot2d(
            axis_off=False,
            cmap="Reds",
            fig_height=None,
            fig_width=6,
            show=True,
            save=False,
            extent="tight",
        )



import warnings

import pytest

from archetypal.simple_glazing import calc_simple_glazing


def test_glazing():
    """Simulates a Double Clear Air Glazing System (Window.exe v.7.5)"""
    res = calc_simple_glazing(0.704, 2.703, 0.786)
    print(res["Thickness"])


def test_absurd():
    """Simulates a Double Clear Air Glazing System (Window.exe v.7.5). Will
    raise an error when checking Visible Transmittance at Normal Incidence +
    Back Side Visible Reflectance at Normal Incidence not <= 1.0"""
    with pytest.warns(UserWarning):
        calc_simple_glazing(0.704, 2.703, 0.9)


@pytest.mark.parametrize(
    "another",
    [
        calc_simple_glazing(0.5, 2.2, 0.21),
        calc_simple_glazing(0.6, 2.2),
        calc_simple_glazing(0.8, 2.2, 0.35),
        calc_simple_glazing(1.2, 0.1, 10),
    ],
)
def test_glazing_unequal(another):
    t1 = calc_simple_glazing(0.6, 2.2, 0.21)
    assert t1 != another


def test_simple_glazing_system_equal():
    dict = calc_simple_glazing(0.6, 2.2, 0.21)
    assert dict["Conductivity"] == 0.11992503503877955


def test_simple_glazing_value_error():
    # Should raise an error since u-value is higher than 7
    with pytest.raises(ValueError):
        calc_simple_glazing(1.2, 8, 10)



import pandas as pd
from path import Path

from archetypal import config, parallel_process, IDF

config(cache_folder="../../tests/.temp/cache", use_cache=True, log_console=True)


def main():

    # setup directories and input files
    necb_basedir = Path("../../tests/input_data/necb")
    files = necb_basedir.glob("Ref*.idf")
    epw = Path("../../tests/input_data/CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw")

    idfs = pd.DataFrame({"file": files, "name": [file.basename() for file in files]})

    # setup the runner. We'll use the DataFrame index as keys (k).
    rundict = {
        k: dict(
            eplus_file=str(file),
            prep_outputs=True,
            weather_file=str(epw),
            expandobjects=False,
            verbose="v",
            design_day=True,
        )
        for k, file in idfs.file.to_dict().items()
    }

    idfs = parallel_process(rundict, IDF, processors=-1, use_kwargs=True)
    return idfs


if __name__ == "__main__":
    config(use_cache=True, log_console=True)
    idfs = main()
    print(idfs)



################################################################################
# Module: simple_glazing.py
# Description: Python implementation of the EnergyPlus Simple Window Model
# License: MIT, see full license in LICENSE.txt
# Web: https://github.com/samuelduchesne/archetypal
################################################################################

import logging as lg
import math
import warnings

import numpy as np

from archetypal import log


def calc_simple_glazing(shgc, u_factor, visible_transmittance=None):
    """Calculates the physical properties of an equivalent single pane of glass
    from the Solar Heat Gain Coefficient, the U-value and the Visible
    Transmittance.

    Args:
        shgc (double): The window's Solar Heat Gain Coefficient
        u_factor (double): The window's U-value
        visible_transmittance (double, optional): The window's visible
            transmittance. If none, the visible transmittance defaults to the
            solar transmittance t_sol.

    Returns:
        dict: A dictionary of properties for the simple glazing system.

    Hint:
        This is an implementation of the `Simple Window Model
        <https://bigladdersoftware.com/epx/docs/8-9/engineering-reference
        /window-calculation-module.html#simple-window-model>`_.

    """

    if isinstance(shgc, str):
        shgc = float(shgc)
    if isinstance(u_factor, str):
        u_factor = float(u_factor)
    if isinstance(visible_transmittance, str):
        try:
            visible_transmittance = float(visible_transmittance)
        except ValueError:
            visible_transmittance = None

    # Limits
    if u_factor > 7:
        raise ValueError(
            "The model cannot support glazing systems with a U higher than "
            "7.0 because  the thermal resistance of the film coefficients "
            "alone can provide this level of performance and  none of the "
            "various resistances can be negative."
        )

    dict = {}

    # Step 1. Determine glass-to-glass Resistance.

    # glass-to-glass resistance (R_l_w) is calculated using:
    R_l_w = r_l_w(u_factor)

    # Step 2. Determine Layer Thickness

    # The thickness of the equivalent layer in units of meters is calculated
    # using:
    Thickness = thickness(R_l_w)

    # Step 3. Determine Layer Thermal Conductivity

    # The effective thermal conductivity, `lambda_eff`, of the equivalent
    # layer is calculated using:
    Lambda_eff = lambda_eff(Thickness, R_l_w)

    # Step 4. Determine Layer Solar Transmittance

    # The layer’s solar transmittance at normal incidence, `T_sol`,
    # is calculated using correlations that are a function of SHGC and U-Factor.

    T_sol = t_sol(shgc, u_factor)

    # Step 5. Determine Layer Solar Reflectance

    R_i_s = r_i_s(shgc, T_sol, u_factor)
    R_o_s = r_o_s(shgc, T_sol, u_factor)
    R_s_f = r_s_f(T_sol, shgc, R_o_s, R_l_w, R_i_s)
    R_s_b = R_s_f  # Both are taken equal

    # The thermal absorptance, or emittance, is taken as 0.84 for both the
    # front and back and the longwave transmittance is 0.0

    # Step 6. Determine Layer Visible Properties
    # The user has the option of entering a value for visible transmittance
    # as one of the simple performance indices.
    # If the user does not enter a value, then the visible properties are the
    # same as the solar properties.
    if visible_transmittance:
        T_vis = visible_transmittance
    else:
        T_vis = T_sol

    R_vis_b = r_vis_b(T_vis)
    R_vis_f = r_vis_f(T_vis)

    # sanity checks
    if T_vis + R_vis_f >= 1.0:
        warnings.warn("T_vis + R_vis_f > 1", UserWarning)
        T_vis -= (T_vis + R_vis_f - 1) * 1.1
    if T_vis + R_vis_b >= 1.0:
        warnings.warn("T_vis + R_vis_b > 1", UserWarning)
        T_vis -= (T_vis + R_vis_b - 1) * 1.1

    # Last Step. Saving results to dict
    dict["SolarHeatGainCoefficient"] = shgc
    dict["UFactor"] = u_factor
    dict["Conductivity"] = Lambda_eff
    dict["Thickness"] = Thickness
    dict["SolarTransmittance"] = T_sol
    dict["SolarReflectanceFront"] = R_s_f
    dict["SolarReflectanceBack"] = R_s_b
    dict["IRTransmittance"] = 0.0
    dict["VisibleTransmittance"] = T_vis
    dict["VisibleReflectanceFront"] = R_vis_f
    dict["VisibleReflectanceBack"] = R_vis_b
    dict["IREmissivityFront"] = 0.84
    dict["IREmissivityBack"] = 0.84
    dict["DirtFactor"] = 1.0  # Clean glass

    dict["Cost"] = 0
    dict["Density"] = 2500
    dict["EmbodiedCarbon"] = 0
    dict["EmbodiedCarbonStdDev"] = 0
    dict["EmbodiedEnergy"] = 0
    dict["EmbodiedEnergyStdDev"] = 0
    dict["Life"] = 1
    dict["SubstitutionRatePattern"] = [1.0]
    dict["SubstitutionTimestep"] = 0
    dict["TransportCarbon"] = 0
    dict["TransportDistance"] = 0
    dict["TransportEnergy"] = 0
    dict["Type"] = "Uncoated"  # TODO Further investigation necessary

    dict["Comments"] = (
        "Properties calculated from Simple Glazing System with "
        "SHGC={:.3f}, UFactor={:.3f} and Tvis={:.3f}".format(shgc, u_factor, T_vis)
    )

    return dict


# region Step 1. Determine glass-to-glass Resistance
def r_i_w(u_factor):
    """The resistance of the interior film coefficient under standard winter
    conditions in units of m\ :sup:`2`\ K/W

    Args:
        u_factor: The U-value of the window including interior and exterior
        surface heat transfer coefficients

    Returns:
        The resistance of the interior film coefficient under standard winter
        conditions

    """
    if u_factor < 5.85:
        return 1 / (0.359073 * math.log(u_factor) + 6.949915)
    return 1 / (1.788041 * u_factor - 2.886625)


def r_o_w(u_factor):
    """The resistance of the exterior film coefficient under standard winter
    conditions in units of m\ :sup:`2`\ K/W

    Args:
        u_factor: The U-value of the window including interior and exterior
        surface heat transfer coefficients

    Returns:
        The resistance of the exterior film coefficient under standard winter
        conditions in units of

    """
    return 1 / (0.025342 * u_factor + 29.162853)


def r_l_w(u_factor):
    """The resisance of the bare window under winter conditions (without the
    film coefficients) in units of
    m\ :sup:`2`\ K/W.

    Args:
        u_factor: The U-value of the window including including interior and
        exterior surface heat transfer
        coefficients.

    Returns:
        The resisance of the bare window under winter conditions (without the
        film coefficients) in units of
        m\ :sup:`2`\ K/W.

    Warnings:
        The model cannot support glazing systems with a U-value higher than
        7.0 because the thermal resistance of
        the film coefficients alone can provide this level of performance and
        none of the various resistances
        can be negative.

    """
    if u_factor > 7.0:
        log(
            "The model cannot support glazing systems with a U-value higher "
            "than 7.0 because the thermal resistance "
            "of the film coefficients alone can provide this level of "
            "performance and none of the various resistances"
            "can be negative.",
            lg.WARNING,
        )
        pass
    return (1 / u_factor) - r_i_w(u_factor) - r_o_w(u_factor)


# endregion


# region Step 2. Determine Layer Thickness
def thickness(r_l_w):
    """The thickness of the equivalent layer in units of meters.

    Args:
        r_l_w (double): The resistance of the bare window under winter
            conditions (without the film coefficients) in units of m\ :sup:`2`\ K`/W.

    Returns:
        double: The thickness of the equivalent layer in units of meters

    """
    if 1 / r_l_w > 7.0:
        return 0.002
    return 0.05914 - 0.00714 / r_l_w


def lambda_eff(thickness, r_l_w):
    """The effective thermal conductivity of the equivalent layer in W/m-K.

    Args:
        thickness: The thickness of the equivalent layer in units of meters.
        r_l_w: The resisance of the bare window under winter conditions (
        without the film coefficients) in
            units of m\ :sup:`2`\ K/W.

    Returns:
        The effective thermal conductivity, λ\ :sub:`eff`, of the equivalent
        layer.

    """
    return thickness / r_l_w


# endregion

# region Step 4. Determine Layer Solar Transmittance
def t_sol_intermediate(shgc, u_factor):
    if u_factor >= 4.5 and shgc < 0.7206:
        return 0.939998 * shgc ** 2 + 0.20332 * shgc
    if u_factor >= 4.5 and shgc >= 0.7206:
        return 1.30415 * shgc - 0.30515
    if u_factor <= 3.4 and shgc <= 0.15:
        return 0.41040 * shgc
    if u_factor <= 3.4 and shgc > 0.15:
        return 0.085775 * shgc ** 2 + 0.963954 * shgc - 0.084958
    pass


def t_sol(shgc, u_factor):
    """The Solat Transmittance at normal incidence t_sol, is calculated using
    correlations
    that are a function of SHGC and U-Factor.

    Warnings:
        for U-values between 3.4 and 4.5, the value for T_sol is interpolated
        using results of the equations for both ranges.

    Args:
        shgc: The window's Solar Heat Gain Coefficient
        u_factor: The window's U-value

    Returns:
        The Solat Transmittance at normal incidence

    """
    if 3.4 <= u_factor <= 4.5:
        return np.interp(
            u_factor,
            [3.4, 4.5],
            [t_sol_intermediate(shgc, 3.4), t_sol_intermediate(shgc, 4.5)],
        )
    return t_sol_intermediate(shgc, u_factor)


# endregion

# region Step 5. Determine Layer Solar Reflectance
def r_i_s_intermediate(shgc, t_sol, u_factor):
    if u_factor >= 4.5:
        return 1 / (
            29.436546 * (shgc - t_sol) ** 3
            - 21.943415 * (shgc - t_sol) ** 2
            + 9.945872 * (shgc - t_sol)
            + 7.426151
        )
    if u_factor <= 3.4:
        return 1 / (
            199.8208128 * (shgc - t_sol) ** 3
            - 90.639733 * (shgc - t_sol) ** 2
            + 19.737055 * (shgc - t_sol)
            + 6.766575
        )


def r_i_s(shgc, t_sol, u_factor):
    """Resistance of the inside coefficient under summer conditions

    Args:
        shgc: The Solar Heat Gain Coefficient
        t_sol: The Solat Transmittance at normal incidence
        u_factor: The window's U-value

    Returns:
        The Resistance of the inside coefficient under summer conditions

    """
    if 3.4 <= u_factor <= 4.5:
        return np.interp(
            u_factor,
            [3.4, 4.5],
            [
                r_i_s_intermediate(shgc, t_sol, 3.4),
                r_i_s_intermediate(shgc, t_sol, 4.5),
            ],
        )
    return r_i_s_intermediate(shgc, t_sol, u_factor)


def r_o_s_intermediate(shgc, t_sol, u_factor):
    if u_factor >= 4.5:
        return 1 / (2.225824 * (shgc - t_sol) + 20.57708)
    if u_factor <= 3.4:
        return 1 / (5.763355 * (shgc - t_sol) + 20.541528)


def r_o_s(shgc, t_sol, u_factor):
    """Resistance of the outside coefficient under summer conditions

    Args:
        shgc: The window's Solar Heat Gain Coefficient
        t_sol: The Solat Transmittance at normal incidence
        u_factor: The window's U-value

    Returns:
        The resistance of the outside coefficient under summer conditions

    """
    if 3.4 <= u_factor <= 4.5:
        return np.interp(
            u_factor,
            [3.4, 4.5],
            [
                r_o_s_intermediate(shgc, t_sol, 3.4),
                r_o_s_intermediate(shgc, t_sol, 4.5),
            ],
        )
    return r_o_s_intermediate(shgc, t_sol, u_factor)


def frac_inward(r_o_s, r_l_w, r_i_s):
    """The inward flowing fraction

    Args:
        r_o_s: Resistance of the inside coefficient under summer conditions
        r_l_w: The resisance of the bare window under winter conditions (
        without the film coefficients) in
            units of m\ :sup:`2`\ K/W
        r_i_s: Resistance of the inside coefficient under summer conditions

    Returns:
        The inward flowing fraction

    """
    return (r_o_s + 0.5 * r_l_w) / (r_o_s + r_l_w + r_i_s)


def r_s_f(t_sol, shgc, r_o_s, r_l_w, r_i_s):
    """The solar reflectance of the front face

    Notes: The solar reflectance of the back face is the same as the solar
    reflectance of the front face.

    Args:
        t_sol (double): The Solat Transmittance at normal incidence.
        shgc (double): The window's Solar Heat Gain Coefficient
        r_o_s (double): Resistance of the outside coefficient under summer
            conditions.
        r_l_w (double): The resisance of the bare window under winter
            conditions (without the film coefficients) in units of m\ :sup:`2`\ K/W.
        r_i_s (double): Resistance of the inside coefficient under summer
        conditions.

    Returns:
        double: Solar reflectance of the front face

    """
    return 1 - t_sol - (shgc - t_sol) / frac_inward(r_o_s, r_l_w, r_i_s)


# endregion


# region Step 6. Determine Layer Visible Properties
def r_vis_b(t_vis):
    """The visible light reflectance for the back surface.

    Args:
        t_vis: The visible transmittance

    Returns:
        double: The visible light reflectance for the back surface
    """
    return -0.7409 * t_vis ** 3 + 1.6531 * t_vis ** 2 - 1.2299 * t_vis + 0.4547


def r_vis_f(t_vis):
    """The visible light reflectance for the front surface

    Args:
        t_vis (double): The visible transmittance

    Returns:
        double: The visible light reflectance for the front surface

    """
    return -0.0622 * t_vis ** 3 + 0.4277 * t_vis ** 2 - 0.4169 * t_vis + 0.2399


# endregion



"""EnergyPlusVersion module."""

import platform
import re
import warnings
from itertools import chain

from packaging.version import Version
from path import Path

from archetypal import settings
from archetypal.eplus_interface.exceptions import InvalidEnergyPlusVersion
from archetypal.settings import ep_version


def get_eplus_dirs(version=ep_version):
    """Return EnergyPlus root folder for a specific version.

    Args:
        version (str): Version number in the form "9-2-0" to search for.

    Returns:
        (Path): The folder path.
    """
    from eppy.runner.run_functions import install_paths

    eplus_exe, eplus_weather = install_paths(version)
    return Path(eplus_exe).dirname()


def get_eplus_basedirs():
    """Return a list of possible E+ install paths."""
    if platform.system() == "Windows":
        eplus_homes = Path("C:\\").dirs("EnergyPlus*")
        return eplus_homes
    elif platform.system() == "Linux":
        eplus_homes = Path("/usr/local/").dirs("EnergyPlus*")
        return eplus_homes
    elif platform.system() == "Darwin":
        eplus_homes = Path("/Applications").dirs("EnergyPlus*")
        return eplus_homes
    else:
        warnings.warn(
            "Archetypal is not compatible with %s. It is only compatible "
            "with Windows, Linux or MacOs" % platform.system()
        )


def _latest_energyplus_version():
    """Find all EnergyPlus installs. and returns the latest version number.

    Only looks in default locations on all platforms.

    Returns:
        (EnergyPlusVersion): The version number of the latest E+ install
    """
    eplus_homes = get_eplus_basedirs()

    # check if any EnergyPlus install exists
    if not eplus_homes:
        raise Exception(
            "No EnergyPlus installation found. Make sure you have EnergyPlus "
            "installed. Go to https://energyplus.net/downloads to download the "
            "latest version of EnergyPlus."
        )

    # Find the most recent version of EnergyPlus installed from the version
    # number (at the end of the folder name)
    return sorted(
        (re.search(r"([\d])-([\d])-([\d])", home.stem).group() for home in eplus_homes),
        reverse=True,
    )[0]


def warn_if_not_compatible():
    """Check if an EnergyPlus install is detected.

    Warnings:
        If the latest version detected is higher than the one specified by
        archetypal, a warning is raised.
    """
    eplus_homes = get_eplus_basedirs()

    if not eplus_homes:
        warnings.warn(
            "No installation of EnergyPlus could be detected on this "
            "machine. Please install EnergyPlus from https://energyplus.net before "
            "using archetypal"
        )


class EnergyPlusVersion(Version):
    """EnergyPlusVersion class.

    This class subclasses the :class:`packaging.version.Version` class. It is usuful
    to compare version numbers together.

    Any EnergyPlusVersion numbers are checked against valid versions before they can
    be initialized.

    Examples:
        To create a version number:

        >>> from archetypal import EnergyPlusVersion
        >>> EnergyPlusVersion("9.2.0")
        <EnergyPlusVersion('9.2.0')>

        An invalid version number raises an exception:

        >>> from archetypal import EnergyPlusVersion
        >>> EnergyPlusVersion("3.2.0")  # "3.2.0" was never released.
        archetypal.eplus_interface.exceptions.InvalidEnergyPlusVersion

    """

    def __init__(self, version):
        """Initialize an EnergyPlusVersion from a version number.

        Args:
            version (str, EnergyPlusVersion): The version number to create. Can be a
                string a tuple or another EnergyPlusVersion object.

        Raises:
            InvalidEnergyPlusVersion: If the version is not a valid version number.
        """
        if isinstance(version, tuple):
            version = ".".join(map(str, version[0:3]))
        if isinstance(version, Version):
            version = ".".join(map(str, (version.major, version.minor, version.micro)))
        if isinstance(version, str) and "-" in version:
            version = version.replace("-", ".")
        super(EnergyPlusVersion, self).__init__(version)
        if self.dash not in self.valid_versions:
            raise InvalidEnergyPlusVersion

    def __repr__(self) -> str:
        """Return the string representation of the object."""
        return "<EnergyPlusVersion({0})>".format(repr(str(self)))

    @classmethod
    def latest(cls):
        """Return the latest EnergyPlus version installed."""
        version = _latest_energyplus_version()
        return cls(version)

    @classmethod
    def current(cls):
        """Return the current EnergyPlus version specified by the main module.

        Specified by :ref:`archetypal.settings.ep_version`
        """
        version = settings.ep_version
        return cls(version)

    @property
    def tuple(self) -> tuple:
        """Return the object as a tuple: (major, minor, micro)."""
        return self.major, self.minor, self.micro

    @property
    def dash(self) -> str:
        """Return the object as a dash-separated string: "major-minor-micro"."""
        return "-".join(map(str, (self.major, self.minor, self.micro)))

    @property
    def valid_versions(self) -> list:
        """List the idd versions installed on this machine."""
        try:
            basedirs_ = []
            for basedir in get_eplus_basedirs():
                updater_ = basedir / "PreProcess" / "IDFVersionUpdater"
                if updater_.exists():
                    basedirs_.append(updater_.files("*.idd"))
            iddnames = set(chain.from_iterable(basedirs_))
        except FileNotFoundError:
            _choices = ["9-2-0"]  # Little hack in case E+ is not installed
        else:
            _choices = set(
                re.match(r"V(.*)-Energy\+", idd.stem).groups()[0] for idd in iddnames
            )
        return _choices



import functools
import time

import numpy as np
import pandas as pd

from archetypal import log


class TabularData(pd.DataFrame):
    """This class serves as a subclass of a pandas DataFrame allowing to add
    additional functionnality
    """

    ARCHETYPE = "Archetype"
    TABULARDATAINDEX = "TabularDataIndex"
    VALUE = "Value"
    REPORTNAME = "ReportName"
    REPORTFORSTRING = "ReportForString"
    TABLENAME = "TableName"
    ROWNAME = "RowName"
    COLUMNNAME = "ColumnName"
    UNITS = "Units"

    @classmethod
    def from_sql(cls, sql_dict):
        """Returns a DataFrame from the 'TabularDataWithStrings' table

        Args:
            sql_dict:
        """

        tab_data_wstring = sql_dict["TabularDataWithStrings"]
        tab_data_wstring.index.names = ["Index"]

        # strip whitespaces
        tab_data_wstring.Value = tab_data_wstring.Value.str.strip()
        tab_data_wstring.RowName = tab_data_wstring.RowName.str.strip()

        return cls(tab_data_wstring)

    @property
    def _constructor(self):
        return TabularData

    @property
    def df(self):
        """Returns the DataFrame of the ReportData"""
        return pd.DataFrame(self)

    def filter_tabular_data(
        self,
        archetype=None,
        tabulardataindex=None,
        value=None,
        reportname=None,
        reportforstring=None,
        tablename=None,
        rowname=None,
        columnname=None,
        units=None,
        inplace=False,
    ):
        """filter RaportData using specific keywords. Each keywords can be a
        tuple of strings (str1, str2, str3) which will return the logical_or on
        the specific column.

        Args:
            archetype (str or tuple):
            tabulardataindex:
            value (str or tuple):
            reportname:
            reportforstring:
            tablename:
            rowname:
            columnname:
            units (str or tuple):
            inplace (str or tuple):

        Returns:
            pandas.DataFrame
        """
        start_time = time.time()
        c_n = []

        if archetype:
            c_1 = (
                conjunction(
                    *[self[self.ARCHETYPE] == archetype for archetype in archetype],
                    logical=np.logical_or
                )
                if isinstance(archetype, tuple)
                else self[self.ARCHETYPE] == archetype
            )
            c_n.append(c_1)
        if tabulardataindex:
            c_2 = (
                conjunction(
                    *[
                        self[self.TABULARDATAINDEX] == tabulardataindex
                        for tabulardataindex in tabulardataindex
                    ],
                    logical=np.logical_or
                )
                if isinstance(tabulardataindex, tuple)
                else self[self.TABULARDATAINDEX] == tabulardataindex
            )
            c_n.append(c_2)
        if value:
            c_3 = (
                conjunction(
                    *[self[self.VALUE] == value for value in value],
                    logical=np.logical_or
                )
                if isinstance(value, tuple)
                else self[self.VALUE] == value
            )
            c_n.append(c_3)
        if reportname:
            c_4 = (
                conjunction(
                    *[self[self.REPORTNAME] == reportname for reportname in reportname],
                    logical=np.logical_or
                )
                if isinstance(reportname, tuple)
                else self[self.REPORTNAME] == reportname
            )
            c_n.append(c_4)
        if value:
            c_5 = (
                conjunction(
                    *[self[self.VALUE] == value for value in value],
                    logical=np.logical_or
                )
                if isinstance(value, tuple)
                else self[self.VALUE] == value
            )
            c_n.append(c_5)
        if reportforstring:
            c_6 = (
                conjunction(
                    *[
                        self[self.REPORTFORSTRING] == reportforstring
                        for reportforstring in reportforstring
                    ],
                    logical=np.logical_or
                )
                if isinstance(reportforstring, tuple)
                else self[self.REPORTFORSTRING] == reportforstring
            )
            c_n.append(c_6)
        if tablename:
            c_7 = (
                conjunction(
                    *[self[self.TABLENAME] == tablename for tablename in tablename],
                    logical=np.logical_or
                )
                if isinstance(tablename, tuple)
                else self[self.TABLENAME] == tablename
            )
            c_n.append(c_7)
        if rowname:
            c_8 = (
                conjunction(
                    *[self[self.ROWNAME] == rowname for rowname in rowname],
                    logical=np.logical_or
                )
                if isinstance(rowname, tuple)
                else self[self.ROWNAME] == rowname
            )
            c_n.append(c_8)
        if columnname:
            c_9 = (
                conjunction(
                    *[self[self.COLUMNNAME] == columnname for columnname in columnname],
                    logical=np.logical_or
                )
                if isinstance(columnname, tuple)
                else self[self.COLUMNNAME] == columnname
            )
            c_n.append(c_9)
        if units:
            c_14 = (
                conjunction(
                    *[self[self.UNITS] == units for units in units],
                    logical=np.logical_or
                )
                if isinstance(units, tuple)
                else self[self.UNITS] == units
            )
            c_n.append(c_14)

        filtered_df = self.loc[conjunction(*c_n, logical=np.logical_and)]
        log("filtered TabularData in {:,.2f} seconds".format(time.time() - start_time))
        if inplace:
            return filtered_df._update_inplace(filtered_df)
        else:
            return filtered_df._constructor(filtered_df).__finalize__(self)


def conjunction(*conditions, logical=np.logical_and):
    """Applies a logical function on n conditons

    Args:
        *conditions:
        logical:
    """
    return functools.reduce(logical, conditions)


def or_conjunction(*conditions):
    """
    Args:
        *conditions:
    """
    return functools.reduce(np.logical_or, conditions)
